'''
Created on Sep 12, 2014

@author: Ronald Haentjens Dekker
'''
import unittest
from collatex import Collation, collate


class Test(unittest.TestCase):

    def test_hermans_witness_order_independence_case_two_witnesses(self):
        collation = Collation()
        collation.add_plain_witness("A", "a b c d F g h i ! K ! q r s t")
        collation.add_plain_witness("B", "a b c d F g h i ! q r s t")
        alignment_table = collate(collation)
        self.assertEquals(["a b c d F g h i!", "K!", "q r s t"], alignment_table.rows[0].to_list())
        self.assertEquals(["a b c d F g h i!", "-", "q r s t"], alignment_table.rows[1].to_list())

    def test_hermans_witness_order_independence_case_three_witnesses(self):
        collation = Collation()
        collation.add_plain_witness("A", "a b c d F g h i ! K ! q r s t")
        collation.add_plain_witness("B", "a b c d F g h i ! q r s t")
        collation.add_plain_witness("C", "a b c d E g h i ! q r s t")
        alignment_table = collate(collation)
        self.assertEquals(["a b c d", "F", "g h i", "! K", "! q r s t"], alignment_table.rows[0].to_list())
        self.assertEquals(["a b c d", "F", "g h i", "-", "! q r s t"], alignment_table.rows[1].to_list())
        self.assertEquals(["a b c d", "E", "g h i", "-", "! q r s t"], alignment_table.rows[2].to_list())

    def test_witness_order(self):
        collation = Collation()
        collation.add_plain_witness("A", "x a y")
        collation.add_plain_witness("B", "x b y")
        collation.add_plain_witness("C", "x a b y")
        alignment_table = collate(collation)
        self.assertEquals(["x", "a", "-", "y"], alignment_table.rows[0].to_list())
        self.assertEquals(["x", "-", "b", "y"], alignment_table.rows[1].to_list())
        self.assertEquals(["x", "a", "b", "y"], alignment_table.rows[2].to_list())

    # TODO: test with x b a y
    
if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
Created on Sep 29, 2014

@author: ronald
'''
import unittest
from collatex import Collation, collate
from collatex.edit_graph_aligner import EditGraphAligner
from collatex.core_classes import VariantGraph


class Test(unittest.TestCase):

    def test_longer_example(self):
        collation = Collation()
        collation.add_plain_witness("A", "The quick brown fox jumps over the dog.")
        collation.add_plain_witness("B", "The brown fox jumps over the lazy dog.")
        collate(collation)

#     # we need to introduce a gap here
    def testOmission(self):
        collation = Collation()
        collation.add_plain_witness("A", "a b c")
        collation.add_plain_witness("B", "b c")
        aligner = EditGraphAligner(collation)
        graph = VariantGraph()
        aligner.collate(graph, collation)
        table = aligner.table
        aligner._debug_edit_graph_table(table)
        self.assertEqual(0, table[0][0].g)
        self.assertEqual(-1, table[0][1].g)
        self.assertEqual(-2, table[0][2].g)
        self.assertEqual(-3, table[0][3].g)
        self.assertEqual(-1, table[1][0].g)
        self.assertEqual(-2, table[1][1].g)
        self.assertEqual(-1, table[1][2].g)
        self.assertEqual(-2, table[1][3].g)
        self.assertEqual(-2, table[2][0].g)
        self.assertEqual(-3, table[2][1].g)
        self.assertEqual(-2, table[2][2].g)
        self.assertEqual(-1, table[2][3].g)


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
Created on Nov 20, 2014

@author: Tara L Andrews

This module contains the exception classes used elsewhere in CollateX.
'''

class TokenError(Exception):
    pass

class UnsupportedError(Exception):
    pass


'''
Created on Aug 4, 2014

@author: ronald
'''
import unittest
from tests import unit_disabled
from collatex.core_classes import Witness
from collatex.experimental_astar_aligner import Aligner


class Test(unittest.TestCase):

    @unit_disabled
    def testOmission(self):
        a = Witness("A", "a b c")
        b = Witness("B", "b c")
        aligner = Aligner(a, b)
        aligner.align()
        
        pass


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testOmission']
    unittest.main()


'''
Created on Jun 6, 2014

@author: Ronald Haentjens Dekker
'''
import unittest
from collatex import Collation
from collatex import collate


class Test(unittest.TestCase):

    #TODO: last one only works when transposition detection is added.
    def testBeckett(self):
        collation = Collation()
        collation.add_plain_witness("1", "The same clock as when for example Magee once died.")
        collation.add_plain_witness("2", "The same as when for example Magee once died.")
        table = collate(collation)
        self.assertEquals(["The same", "clock", "as when for example Magee once died."], table.rows[0].to_list())
        self.assertEquals(["The same", "-", "as when for example Magee once died."], table.rows[1].to_list())
        
#         table.print_plain_text()
#         "The same as when for example McKee once died .",//
#         "The same as when among others Darly once died & left him.",//
#       #  "The same as when Darly among others once died and left him.");

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
Created on May 3, 2014

@author: Ronald Haentjens Dekker
'''
from collatex.core_classes import VariantGraph, Witness, join, AlignmentTable, Row, WordPunctuationTokenizer
from collatex.collatex_suffix import ExtendedSuffixArray
from collatex.exceptions import UnsupportedError
from collatex.linsuffarr import SuffixArray, UNIT_BYTE
from ClusterShell.RangeSet import RangeSet
import json
from collatex.edit_graph_aligner import EditGraphAligner
from collatex.display_module import display_alignment_table_as_HTML
from collatex.display_module import display_variant_graph_as_SVG

# Valid options for output are:
# "table" for the alignment table (default)
# "graph" for the variant graph
# "json" for the alignment table exported as JSON
def collate(collation, output="table", layout="horizontal", segmentation=True, near_match=False, astar=False, debug_scores=False):
    algorithm = EditGraphAligner(collation, near_match=near_match, astar=astar, debug_scores=debug_scores)
    # build graph
    graph = VariantGraph()
    algorithm.collate(graph, collation)
    # join parallel segments
    if segmentation:
        join(graph)
    # check which output format is requested: graph or table
    if output == "svg":
        return display_variant_graph_as_SVG(graph)
    if output=="graph": 
        return graph
    # create alignment table
    table = AlignmentTable(collation, graph, layout)
    if output == "json":
        return export_alignment_table_as_json(table)
    if output == "html":
        return display_alignment_table_as_HTML(table)
    if output == "table":
        return table
    else:
        raise Exception("Unknown output type: "+output)
    


#TODO: this only works with a table output at the moment
#TODO: store the tokens on the graph instead
def collate_pretokenized_json(json, output='table', layout='horizontal', **kwargs):
    # Takes more or less the same arguments as collate() above, but with some restrictions.
    # Only output types 'json' and 'table' are supported.
    if output not in ['json', 'table']:
        raise UnsupportedError("Output type" + kwargs['output'] + "not supported for pretokenized collation")
    if 'segmentation' in kwargs and kwargs['segmentation']:
        raise UnsupportedError("Segmented output not supported for pretokenized collation")
    kwargs['segmentation'] = False

    # For each witness given, make a 'shadow' witness based on the normalization tokens
    # that will actually be collated.
    tokenized_witnesses = []
    collation = Collation()
    for witness in json["witnesses"]:
        collation.add_witness(witness)
        tokenized_witnesses.append(witness["tokens"])
    at = collate(collation, output="table", **kwargs)
    tokenized_at = AlignmentTable(collation, layout=layout)
    for row, tokenized_witness in zip(at.rows, tokenized_witnesses):
        new_row = Row(row.header)
        tokenized_at.rows.append(new_row)
        token_counter = 0
        for cell in row.cells:
            if cell != "-":
                new_row.cells.append(tokenized_witness[token_counter])
                token_counter+=1
            else:
                #TODO: should probably be null or None instead, but that would break the rendering at the moment 
                new_row.cells.append({"t":"-"})
    if output=="json":
        return export_alignment_table_as_json(tokenized_at)
    if output=="table":
        # transform JSON objects to "t" form.
        for row in tokenized_at.rows:
            row.cells = [cell["t"]  for cell in row.cells]
        return tokenized_at

def export_alignment_table_as_json(table, indent=None, status=False):
    json_output = {}
    json_output["table"]=[]
    sigli = []
    for row in table.rows:
        sigli.append(row.header)
        json_output["table"].append([[cell] for cell in row.cells])
    json_output["witnesses"]=sigli
    if status:
        variant_status = []
        for column in table.columns:
            variant_status.append(column.variant)
        json_output["status"]=variant_status
    return json.dumps(json_output, sort_keys=True, indent=indent)

'''
Suffix specific implementation of Collation object
'''
class Collation(object):

    @classmethod
    def create_from_dict(cls, data, limit=None):
        witnesses = data["witnesses"]
        collation = Collation()
        for witness in witnesses[:limit]:
            # generate collation object from json_data
            collation.add_witness(witness)
        return collation

    @classmethod
    # json_data can be a string or a file
    def create_from_json(cls, json_data):
        data = json.load(json_data)
        collation = cls.create_from_dict(data)
        return collation

    def __init__(self):
        self.witnesses = []
        self.counter = 0
        self.witness_ranges = {}
        self.combined_string = ""
        self.cached_suffix_array = None

    # the tokenization process happens multiple times
    # and by different tokenizers. This should be fixed
    def add_witness(self, witnessdata):
        # clear the suffix array and LCP array cache
        self.cached_suffix_array = None
        witness = Witness(witnessdata)
        self.witnesses.append(witness)
        witness_range = RangeSet()
        witness_range.add_range(self.counter, self.counter+len(witness.tokens()))
        # the extra one is for the marker token
        self.counter += len(witness.tokens()) +2 # $ + number 
        self.witness_ranges[witness.sigil] = witness_range
        if not self.combined_string == "":
            self.combined_string += " $"+str(len(self.witnesses)-1)+ " "
        self.combined_string += witness.content

    def add_plain_witness(self, sigil, content):
        return self.add_witness({'id':sigil, 'content':content})

    def get_range_for_witness(self, witness_sigil):
        if not witness_sigil in self.witness_ranges:
            raise Exception("Witness "+witness_sigil+" is not added to the collation!")
        return self.witness_ranges[witness_sigil]

    def get_combined_string(self):
        return self.combined_string

    def get_sa(self):
        #NOTE: implemented in a lazy manner, since calculation of the Suffix Array and LCP Array takes time
        if not self.cached_suffix_array:
            # Unit byte is done to skip tokenization in third party library
            self.cached_suffix_array = SuffixArray(self.tokens, unit=UNIT_BYTE)
        return self.cached_suffix_array

    def get_suffix_array(self):
        sa = self.get_sa()
        return sa.SA

    def get_lcp_array(self):
        sa = self.get_sa()
        return sa._LCP_values


    def to_extended_suffix_array(self):
        return ExtendedSuffixArray(self.tokens, self.get_suffix_array(), self.get_lcp_array())

    @property
    def tokens(self):
        #print("COLLATION TOKENIZE IS CALLED!")
        #TODO: complete set of witnesses is retokenized here!
        tokenizer = WordPunctuationTokenizer()
        tokens = tokenizer.tokenize(self.get_combined_string())
        return tokens






'''
Created on Jun 11, 2014

@author: Ronald Haentjens Dekker
'''
import unittest
from collatex.core_functions import Collation, collate


class Test(unittest.TestCase):

    def testPlainTableRendering(self):
        collation = Collation()
        collation.add_plain_witness("A", "This very quick very quick brown wombat")
        collation.add_plain_witness("B", "That very quick brown koala")
        collation.add_plain_witness("C", "That very quick brown kangaroo")
        expected_output = """+---+-----------------+------------------+----------+
| A | This very quick | very quick brown | wombat   |
| B | That            | very quick brown | koala    |
| C | That            | very quick brown | kangaroo |
+---+-----------------+------------------+----------+"""
        plain_text_output = str(collate(collation))
        self.assertEquals(expected_output, plain_text_output)

    def testPlainTableRenderingNoParallelSegmentation(self):
        collation = Collation()
        collation.add_plain_witness("A", "This very quick very quick brown wombat")
        collation.add_plain_witness("B", "That very quick brown koala")
        collation.add_plain_witness("C", "That very quick brown kangaroo")
        expected_output = """\
+---+------+------+-------+------+-------+-------+----------+
| A | This | very | quick | very | quick | brown | wombat   |
| B | That | -    | -     | very | quick | brown | koala    |
| C | That | -    | -     | very | quick | brown | kangaroo |
+---+------+------+-------+------+-------+-------+----------+"""
        plain_text_output = str(collate(collation, segmentation=False))
        self.assertEquals(expected_output, plain_text_output)

    def testPlainTableRenderingVertical(self):
        collation = Collation()
        collation.add_plain_witness("A", "This very quick very quick brown wombat")
        collation.add_plain_witness("B", "That very quick brown koala")
        collation.add_plain_witness("C", "That very quick brown kangaroo")
        expected_output = """\
+------------------+------------------+------------------+
|        A         |        B         |        C         |
+------------------+------------------+------------------+
| This very quick  |       That       |       That       |
+------------------+------------------+------------------+
| very quick brown | very quick brown | very quick brown |
+------------------+------------------+------------------+
|      wombat      |      koala       |     kangaroo     |
+------------------+------------------+------------------+"""
        plain_text_output = str(collate(collation, layout="vertical"))
        self.assertEquals(expected_output, plain_text_output)

    def testJSONAlignmentTableRendering(self):
        collation = Collation()
        collation.add_plain_witness("A", "This very quick very quick brown wombat")
        collation.add_plain_witness("B", "That very quick brown koala")
        collation.add_plain_witness("C", "That very quick brown kangaroo")
        expected_output = '{"table": [[["This very quick"], ["very quick brown"], ["wombat"]], [["That"], ["very quick brown"], ["koala"]], [["That"], ["very quick brown"], ["kangaroo"]]], "witnesses": ["A", "B", "C"]}'
        json = collate(collation, output="json")
        self.assertEquals(expected_output, json)

    def testColumnStatusInAlignmentTable(self):
        collation = Collation()
        collation.add_plain_witness("A", "The quick brown fox jumps over the dog.")
        collation.add_plain_witness("B", "The brown fox jumps over the lazy dog.")
        alignment_table = collate(collation)
        status_array = []
        for column in alignment_table.columns:
            status_array.append(column.variant)
        self.assertEqual([False, True, False, True, False], status_array)
        collation.add_plain_witness("C", "The brown fox walks around the lazy dog.")
        collate(collation)    
        alignment_table = collate(collation)
        status_array = []
        for column in alignment_table.columns:
            status_array.append(column.variant)
        self.assertEqual([False, True, False, True, False, True, False], status_array)
        
if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
Created on Aug 5, 2014

@author: Ronald Haentjens Dekker
'''

from collatex.core_classes import CollationAlgorithm
from collatex.suffix_based_scorer import Scorer
from prettytable import PrettyTable
from collatex.astar import AStarNode, AStar

class AstarEditGraphNode(AStarNode):
    def __init__(self, aligner, y, x):
        self.aligner = aligner
        self.y = y
        self.x = x
        self.match = False
        self.segments = 0 #TODO: remove
        super(AstarEditGraphNode, self).__init__()

    def __repr__(self):
        return repr(self.g)
#         return str(self.y)+" "+str(self.x)+" "+
    
    def is_end_node(self):
        is_end = self.y == self.aligner.length_witness_b and self.x == self.aligner.length_witness_a
#         print("Node: "+str(self.y)+" "+str(self.x)+" "+str(self.edit_operation)+" is end: "+str(is_end))
        return is_end 
    
    # TODO: not nice: scorer already updates global score of other..
    def move_cost(self, other):
        #NOTE: possible optimization: you don't always need to fetch the tokens!
        token_a = self.aligner.tokens_witness_a[other.x-1]
        token_b = self.aligner.tokens_witness_b[other.y-1]
        edit_operation = 1
        if other.x -1 == self.x and other.y - 1 == self.y:
            edit_operation = 0
        #     def score_cell(self, table_node, parent_node, token_a, token_b, y, x, edit_operation):
        self.aligner.scorer.score_cell(other, self, token_a, token_b, other.y, other.x, edit_operation)
        cost = other.g - self.g
#         print("From Node: "+str(self.y)+" "+str(self.x)+" "+str(self.edit_operation)+" to other: "+str(other.y)+" "+str(other.x)+" "+str(other.edit_operation)+" cost: "+str(cost))
        return -cost

class AstarEditGraphAligner(AStar):
    def __init__(self, tokens_witness_a, tokens_witness_b, scorer):
        self.tokens_witness_a = tokens_witness_a
        self.tokens_witness_b = tokens_witness_b
        self.scorer = scorer
        self.table = None #TODO: not nice!
        self.length_witness_a = len(self.tokens_witness_a)
        self.length_witness_b = len(self.tokens_witness_b)

    def create_childnodes(self, node):
        # we create 3 child nodes
        # NOTE: In some cases it is possible to get away with only 2 nodes
        # NOTE: possible performance enhancement
        child_nodes = []
        if node.y < self.length_witness_b:
            child_node = self.table[node.y+1][node.x]
            child_nodes.append(child_node)
        if node.x < self.length_witness_a:
            child_node = self.table[node.y][node.x+1]
            child_nodes.append(child_node)
        if node.y < self.length_witness_b and node.x < self.length_witness_a:
            child_node = self.table[node.y+1][node.x+1]
            child_nodes.append(child_node)
        return child_nodes

    def heuristic(self, node):
        distance_to_end_node_horizontal = (self.length_witness_a-node.x)
        distance_to_end_node_vertical = (self.length_witness_b-node.y)
        gap_penalty = abs(distance_to_end_node_horizontal - distance_to_end_node_vertical)
#         print("heuristic penalty: "+str(node.y)+" "+str(node.x)+" penalty: "+str(-gap_penalty))
        return gap_penalty

class EditGraphNode(object):
    def __init__(self):
        self.g = 0 # global score 
        self.segments = 0 # number of segments
        self.match = False # this node represents a match or not
        
    def __repr__(self):
        return repr(self.g)

    '''
    Aligner based on an edit graph.
    It needs a g function and a definition of a match.
    Since every node of the graph has three children the graph is represented as a table internally.
    Default implementation is a* based.
    '''
class EditGraphAligner(CollationAlgorithm):
    def __init__(self, collation, near_match=False, astar=False, debug_scores=False):
        self.collation = collation
        self.debug_scores = debug_scores
        self.scorer = Scorer(collation, near_match)
        if not astar:
            self.align_function = self._align_table
        else:
            print("INFO: Aligning using a* search algorithm. BETA quality.")
            self.align_function = self._align_astar
            
    def collate(self, graph, collation):
        '''
        :type graph: VariantGraph
        :type collation: Collation
        '''
        # Build the variant graph for the first witness
        # this is easy: generate a vertex for every token
        first_witness = collation.witnesses[0]
        tokens = first_witness.tokens()
        token_to_vertex = self.merge(graph, first_witness.sigil, tokens)
        
        # let the scorer prepare the first witness
        self.scorer.prepare_witness(first_witness)
        
        # construct superbase
        superbase = tokens
        
        # align witness 2 - n
        for x in range(1, len(collation.witnesses)):
            next_witness = collation.witnesses[x]
        
            # let the scorer prepare the next witness
            self.scorer.prepare_witness(next_witness)
            
#             # VOOR CONTROLE!
#             alignment = self._align_table(superbase, next_witness, token_to_vertex)
#             self.table2 = self.table
            
            # alignment = token -> vertex
            alignment = self.align_function(superbase, next_witness, token_to_vertex)
        
            # merge
            token_to_vertex.update(self.merge(graph, next_witness.sigil, next_witness.tokens(), alignment))

#             print("actual")
#             self._debug_edit_graph_table(self.table)
#             print("expected")
#             self._debug_edit_graph_table(self.table2)
            
            # change superbase
            superbase = self.new_superbase
        
        if self.debug_scores:
            self._debug_edit_graph_table(self.table)
        
    def _align_astar(self, superbase, witness, token_to_vertex, control_table=None):
        self.tokens_witness_a = superbase
        self.tokens_witness_b = witness.tokens()
        self.length_witness_a = len(self.tokens_witness_a)
        self.length_witness_b = len(self.tokens_witness_b)
        self.control_table = control_table
        aligner = AstarEditGraphAligner(self.tokens_witness_a, self.tokens_witness_b, self.scorer)
        self.table = [[AstarEditGraphNode(aligner, y, x) for x in range(self.length_witness_a+1)] for y in range(self.length_witness_b+1)]
        aligner.table = self.table
        start = self.table[0][0]
        path = aligner.search(start, self.control_table)
        
        # transform path into an alignment
        alignment = {}

        # segment stuff
        # note we traverse from left to right!
        self.last_x = 0
        self.last_y = 0
        self.new_superbase=[]
        
        for element in path:
#             print(element.y, element.x)
            
            if element.match == True:
                # process segments
                self.newer_add_to_superbase(self.tokens_witness_a, self.tokens_witness_b, element.x, element.y)
                self.last_x = element.x
                self.last_y = element.y
                # process alignment
                token = self.tokens_witness_a[element.x-1]
                token2 = self.tokens_witness_b[element.y-1]
                vertex = token_to_vertex[token]
                alignment[token2] = vertex
                # add match to superbase
                self.new_superbase.append(token)

        # process additions/omissions in the begin of the superbase/witness
        self.newer_add_to_superbase(self.tokens_witness_a, self.tokens_witness_b, self.length_witness_a, self.length_witness_b)
        return alignment
    
    
    def newer_add_to_superbase(self, witness_a, witness_b, x, y):
        if x - self.last_x - 1 > 0 or y - self.last_y - 1 > 0:
            # create new segment
            omitted_base = witness_a[self.last_x:x - 1]
            added_witness = witness_b[self.last_y:y - 1]
            self.new_superbase += omitted_base
            self.new_superbase += added_witness
    
    def _align_table(self, superbase, witness, token_to_vertex):
        self.tokens_witness_a = superbase
        self.tokens_witness_b = witness.tokens()
        self.length_witness_a = len(self.tokens_witness_a)
        self.length_witness_b = len(self.tokens_witness_b)
        self.table = [[EditGraphNode() for _ in range(self.length_witness_a+1)] for _ in range(self.length_witness_b+1)]

        # per diagonal calculate the score (taking into account the three surrounding nodes)
        self.traverse_diagonally()

        alignment = {}

        # segment stuff
        # note we traverse from right to left!
        self.last_x = self.length_witness_a
        self.last_y = self.length_witness_b
        self.new_superbase=[]
        
        # start lower right cell
        x = self.length_witness_a
        y = self.length_witness_b
        # work our way to the upper left
        while x > 0 and y > 0:
            self._process_cell(token_to_vertex, self.tokens_witness_a, self.tokens_witness_b, alignment, x, y)
            # examine neighbor nodes
            nodes_to_examine = set()
            nodes_to_examine.add(self.table[y][x-1])
            nodes_to_examine.add(self.table[y-1][x])
            nodes_to_examine.add(self.table[y-1][x-1])
            # calculate the maximum scoring parent node
            parent_node = max(nodes_to_examine, key=lambda x: x.g)
            # move position
            if self.table[y-1][x-1] == parent_node:
                # another match or replacement
                y = y -1
                x = x -1
            else:
                if self.table[y-1][x] == parent_node:
                    #omission?
                    y = y -1
                else:
                    if self.table[y][x-1] == parent_node:
                        #addition?
                        x = x -1
        # process additions/omissions in the begin of the superbase/witness
        self.add_to_superbase(self.tokens_witness_a, self.tokens_witness_b, 0, 0)
        return alignment
        

    def add_to_superbase(self, witness_a, witness_b, x, y):
#         print self.last_x - x - 1, self.last_y - y - 1
        if self.last_x - x - 1 > 0 or self.last_y - y - 1 > 0:
#             print x, self.last_x, y, self.last_y 
            # create new segment
            omitted_base = witness_a[x:self.last_x - 1]
#             print omitted_base
            added_witness = witness_b[y:self.last_y - 1]
#             print added_witness
            self.new_superbase = added_witness + self.new_superbase
            self.new_superbase = omitted_base + self.new_superbase

    def _process_cell(self, token_to_vertex, witness_a, witness_b, alignment, x, y):
        cell = self.table[y][x]
        # process segments
        if cell.match == True:
            self.add_to_superbase(witness_a, witness_b, x, y)
            self.last_x = x
            self.last_y = y
        # process alignment
        if cell.match == True:
            token = witness_a[x-1]
            token2 = witness_b[y-1]
            vertex = token_to_vertex[token]
            alignment[token2] = vertex
#             print("match")
#             print(token2)
            self.new_superbase.insert(0, token)
        return cell

    # This function traverses the table diagonally and scores each cell.
    # Original function from Mark Byers; translated from C into Python.
    def traverse_diagonally(self):
        m = self.length_witness_b+1
        n = self.length_witness_a+1
        for _slice in range(0, m + n - 1, 1):
            z1 = 0 if _slice < n else _slice - n + 1;
            z2 = 0 if _slice < m else _slice - m + 1;
            j = _slice - z2
            while j >= z1:
                x = _slice - j
                y = j
                self.score_cell(y, x)
                j -= 1

    def score_cell(self, y, x):
        # initialize root node score to zero (no edit operations have
        # been performed)
        if y == 0 and x == 0:
            self.table[y][x].g = 0
            return 
        # examine neighbor nodes
        nodes_to_examine = set()
        # fetch existing score from the left node if possible
        if x > 0:
            nodes_to_examine.add(self.table[y][x-1])
        if y > 0:
            nodes_to_examine.add(self.table[y-1][x])
        if x > 0 and y > 0:
            nodes_to_examine.add(self.table[y-1][x-1])
        # calculate the maximum scoring parent node
        parent_node = max(nodes_to_examine, key=lambda x: x.g)
        if parent_node == self.table[y-1][x-1]:
            edit_operation = 0
        else:
            edit_operation = 1
        token_a = self.tokens_witness_a[x-1]
        token_b = self.tokens_witness_b[y-1]
        self.scorer.score_cell(self.table[y][x], parent_node, token_a, token_b, y, x, edit_operation)

    def _debug_edit_graph_table(self, table):
        # print the table horizontal
        x = PrettyTable()
        x.header=False
        for y in range(0, len(table)):
            cells = table[y]
            x.add_row(cells)
        # alignment can only be set after the field names are known.
        # since add_row sets the field names, it has to be set after x.add_row(cells)
        x.align="l"
        print(x)
        return x
    




# author: by Ronald Haentjens Dekker (created 4 August 2014)

class AStar(object):
        
    def heuristic(self, node):
        raise NotImplementedError
      
    def create_childnodes(self, node):
        raise NotImplementedError
      
    def search(self, start, control_table):
        # TODO: should become priority queue to store nodes
        heap = [] 
        
        # Set to store previously visited nodes
        visited = set()

        # put the initial node on the queue
        heap.append(start)
 
        while heap:
            heap.sort(key=lambda x:x.g+x.h)
#             print(heap)
            node = heap.pop(0)
#             print("selected (lowest): "+str(node.y)+" "+str(node.x))
            if node.is_end_node():
                return self.reconstruct_path(node)
            visited.add(node)
            children = self.create_childnodes(node)
            for child in children:
                if child in visited:
                    continue
                old_g = child.g
                tentative_g_score = node.g + node.move_cost(child)
                child.g = old_g
                if not child in heap or tentative_g_score < child.g:
                    child.parent = node
                    child.g = tentative_g_score
                    child.h = self.heuristic(child)
                    if not child in heap:
                        heap.append(child)
#                     else:
#                         print("Child is already in heap, will be resorted!")
        # entire tree searched, no goal state found
        return None

    def reconstruct_path(self, current):
        path = []
        while current.parent:
            path.append(current)
            current = current.parent
        return path[::-1]

 
class AStarNode(object):
    def __init__(self):
        self.g = 0
        self.h = 0
        self.parent = None
    
    def move_cost(self, other):
        raise NotImplementedError
    
    def is_end_node(self):
        raise NotImplementedError
    


import unittest
from collatex.core_classes import Tokenizer


class Test(unittest.TestCase):

    def test_tokenize(self):
        contents = "a b c"
        tokenizer = Tokenizer()
        #print contents
        self.assertEquals(["a","b","c"], tokenizer.tokenize(contents))
    
if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testSuffix']
    unittest.main()


'''
Created on Apr 7, 2014

@author: Ronald Haentjens Dekker
'''
from ClusterShell.RangeSet import RangeSet

class Stack(list):
    def push(self, item):
        self.append(item)
        
    def peek(self):
        return self[-1]
    
class PartialOverlapException(Exception):
    pass

class ExtendedSuffixArray(object):
    
    def __init__(self, tokens, sa_array, lcp_array):
        self.tokens = tokens
        self.SA = sa_array
        self.LCP = lcp_array

    # NOTE: LCP intervals can 1) ascend or 2) descend or 3) first ascend and then descend. 4) descend, ascend
    def split_lcp_array_into_intervals(self):
        closed_intervals = []
        previous_lcp_value = 0
        open_intervals = Stack()
        for idx in range(0, len(self.LCP)):
            lcp_value = self.LCP[idx]
#             print(lcp_value)
            if lcp_value > previous_lcp_value:
                open_intervals.push((idx-1, lcp_value))
                previous_lcp_value = lcp_value
            if lcp_value < previous_lcp_value:
                # close open intervals that are larger than current lcp_value
                while open_intervals and open_intervals.peek()[1] > lcp_value:
#                     print("Peek: "+str(open_intervals.peek()))
                    (start, length) = open_intervals.pop()
                    #TODO: FIX NUMBER OF SIBLINGS!
                    closed_intervals.append(LCPInterval(self.tokens, self.SA, self.LCP, start, idx-1, length, 0))
#                     print("new: "+repr(closed_intervals[-1]))
                # then: open a new interval starting with start filter open intervals.
                if lcp_value > 0:
                    start = closed_intervals[-1].start
                    open_intervals.push((start, lcp_value))
                previous_lcp_value = lcp_value
        # add all the open intervals to the result
#         print("Closing remaining:")
        for start, length in open_intervals:
            #TODO: FIX NUMBER OF SIBLINGS!
            closed_intervals.append(LCPInterval(self.tokens, self.SA, self.LCP, start, len(self.LCP)-1, length, 0))
#             print("new: "+repr(closed_intervals[-1]))
        return closed_intervals

    def list_prefixes(self):
        for idx in range(0, len(self.LCP)):
            if self.LCP[idx] > 0:
                prefix = " ".join(self.tokens[self.SA[idx]:self.SA[idx]+self.LCP[idx]])
                print(prefix)
            else:
                print("--------")

# parts of the LCP array become potential blocks.
# minimum block_length: the number of tokens a single occurrence of this block spans
# block_occurrences: the ranges within the suffix array that this block spans
class LCPInterval(object):
    
    def __init__(self, tokens, SA, LCP, start, end, length, number_of_siblings):
        self.tokens = tokens
        self.SA = SA
        self.LCP = LCP
        self.start = start
        self.end = end
        self.length = length
        self.number_of_siblings = number_of_siblings
        
    @property
    def minimum_block_length(self):
        return self.length
    
    @property
    def number_of_occurrences(self):
        return self.end - self.start + 1
    
    def block_occurrences(self):
        block_occurrences = []
        for idx in range(self.start, self.end + 1):
            block_occurrences.append(self.SA[idx])
        return block_occurrences

    def info(self):
        return "looking at: "+str(self)
    
    @property            
    def token_start_position(self):
        return min(self.block_occurrences())
    
    def calculate_non_overlapping_range_with(self, occupied):
        # convert block occurrences into ranges
        potential_block_range = RangeSet()
        for occurrence in self.block_occurrences():
            potential_block_range.add_range(occurrence, occurrence + self.minimum_block_length)
        #check the intersection with the already occupied ranges
        block_intersection = potential_block_range.intersection(occupied)
        if not block_intersection:
            # no overlap, return complete block_range
            return potential_block_range
        # There is overlap with occupied range
        # we need to deal with it
        real_block_range = RangeSet()
        for lower in potential_block_range.contiguous():
            # TODO: what I really want here is a find first over a generator
            upper = [x for x in block_intersection.contiguous() if x[0] >= lower[0]]
            if upper:
                lower = lower[0]
                upper = upper[0][0]
                if lower != upper:
                    real_block_range.add_range(lower, upper)
        if not real_block_range:
            # There is complete overlap, so return None
            return None
        # Assert: check that the first slice is not larger than potential block length!
        first_range = next(real_block_range.contiguous())
        if first_range[-1]-first_range[0]+1>self.minimum_block_length:
            raise PartialOverlapException()
        return real_block_range
    
    def show_lcp_array(self):
        return self.LCP[self.start:self.end+1]
        
    def __str__(self):
        part1= "<"+" ".join(self.tokens[self.SA[self.start]:self.SA[self.start]+min(10, self.minimum_block_length)])
        return part1+"> with "+str(self.number_of_occurrences)+" occurrences and length: "+str(self.minimum_block_length)+" and number of siblings: "+str(self.number_of_siblings)
 
    def __repr__(self):
        return "LCPivl: "+str(self.token_start_position)+","+str(self.minimum_block_length)+","+str(self.number_of_occurrences)
            
  

class Block(object):

    def __init__(self, ranges):
        """
        :type ranges: RangeSet
        """
        self.ranges = ranges

    def __hash__(self):
        return hash(self.ranges.__str__())

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False

    def __str__(self):
        return "Block with occurrences "+str(self.ranges)

    def __repr__(self):
        return "Block: "+str(self.ranges)

# Class represents a range within one witness that is associated with a block
class Occurrence(object):

    def __init__(self, token_range, block):
        self.token_range = token_range
        self.block = block

    def __repr__(self):
        return str(self.token_range)

    @property
    def lower_end(self):
        return self.token_range[0]

    def is_in_range(self, position):
        return position in self.token_range

# Class represents a witness which consists of occurrences of blocks            
class BlockWitness(object):

    def __init__(self, occurrences, tokens):
        self.occurrences = occurrences
        self.tokens = tokens

    def debug(self):
        result = []
        for occurrence in self.occurrences:
            result.append(' '.join(self.tokens[occurrence.token_range.slices().next()]))
        return result


    







'''
Created on Sep 12, 2014

@author: Ronald Haentjens Dekker
'''
import unittest
from tests import unit_disabled
from collatex.core_functions import collate_pretokenized_json


class Test(unittest.TestCase):
    json_in = {
      "witnesses" : [
        {
          "id" : "A",
          "tokens" : [
              { "t" : "I", "ref" : 123 },
              { "t" : "bought" , "adj" : True },
              { "t" : "this", "id" : "x3" },
              { "t" : "glass", "id" : "x4" },
              { "t" : ",", "type" : "punct" },
              { "t" : "because", "id" : "x5" },
              { "t" : "it", "id" : "x6" },
              { "t" : "matches" },
              { "t" : "those", "id" : "x7" },
              { "t" : "dinner", "id" : "x8" },
              { "t" : "plates", "id" : "x9" },
              { "t" : ".", "type" : "punct" }
          ]
        },
        {
          "id" : "B",
          "tokens" : [
              { "t" : "I" },
              { "t" : "bought" , "adj" : True },
              { "t" : "those", "id" : "abc" },
              { "t" : "glasses", "id" : "xyz" },
              { "t" : ".", "type" : "punct" }
          ]
        }
      ]
    }

    def test_exact_matching(self):
        result = collate_pretokenized_json(self.json_in)
        self.assertEquals(["I", "bought", "this", "glass", ",", "because", "it", "matches", "those", "dinner", "plates", "."],
                          result.rows[0].to_list())
        self.assertEquals(["I", "bought", "-", "-", "-", "-", "-", "-", "those", "glasses", "-", "."], result.rows[1].to_list())

    def test_near_matching(self):
        result = collate_pretokenized_json(self.json_in, near_match=True)
        self.assertEquals(["I", "bought", "this", "glass", ",", "because", "it", "matches", "those", "dinner", "plates", "."],
                          result.rows[0].to_list())
        self.assertEquals(["I", "bought", "those", "glasses", "-", "-", "-", "-", "-", "-", "-", "."], result.rows[1].to_list())

    # Re-enable this one if segmented output is ever supported on tokenized collation
    @unit_disabled
    def test_near_matching_segmented(self):
        result = collate_pretokenized_json(self.json_in, near_match=True, segmentation=True)
        self.assertEquals(["I bought", "this glass, because it matches those dinner plates."],
                          result.rows[0].to_list())
        self.assertEquals(["I bought", "those glasses."], result.rows[1].to_list())


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testOmission']
    unittest.main()



'''
Created on Jun 11, 2014

@author: Ronald Haentjens Dekker
'''
import unittest
from collatex.core_functions import collate_pretokenized_json
from collatex.exceptions import *
    

class Test(unittest.TestCase):

    def testJSONOutputPretokenizedJSON(self):
        json_in = {
      "witnesses" : [
        {
          "id" : "A",
          "tokens" : [
              { "t" : "A", "ref" : 123 },
              { "t" : "black" , "adj" : True },
              { "t" : "cat", "id" : "xyz" }
          ]
        },
        {
          "id" : "B",
          "tokens" : [
              { "t" : "A" },
              { "t" : "white" , "adj" : True },
              { "t" : "kitten.", "n" : "cat" }
          ]
        }
      ]
    }

        expected_json = '{"table": [[[{"ref": 123, "t": "A"}], [{"adj": true, "t": "black"}], [{"id": "xyz", "t": "cat"}]], [[{"t": "A"}], [{"adj": true, "t": "white"}], [{"n": "cat", "t": "kitten."}]]], "witnesses": ["A", "B"]}'
        json_out = collate_pretokenized_json(json_in, output="json")
        self.assertEqual(expected_json, json_out)

    def testJSONOutput_empty_cells_in_output(self):
        json_in = {
      "witnesses" : [
        {
          "id" : "A",
          "tokens" : [
              { "t" : "A", "ref" : 123 },
              { "t" : "black" , "adj" : True },
              { "t" : "cat", "id" : "xyz" }
          ]
        },
        {
          "id" : "B",
          "tokens" : [
              { "t" : "A" },
              { "t" : "kitten.", "n" : "cat" }
          ]
        }
      ]
    }
        expected_json = '{"table": [[[{"ref": 123, "t": "A"}], [{"adj": true, "t": "black"}], [{"id": "xyz", "t": "cat"}]], [[{"t": "A"}], [{"t": "-"}], [{"n": "cat", "t": "kitten."}]]], "witnesses": ["A", "B"]}'
        json_out = collate_pretokenized_json(json_in, output="json")
        self.assertEqual(expected_json, json_out)

    def testHTMLOutputPretokenizedJSON(self):
        json_in = {
      "witnesses" : [
        {
          "id" : "A",
          "tokens" : [
              { "t" : "A", "ref" : 123 },
              { "t" : "black" , "adj" : True },
              { "t" : "cat", "id" : "xyz" }
          ]
        },
        {
          "id" : "B",
          "tokens" : [
              { "t" : "A" },
              { "t" : "white" , "adj" : True },
              { "t" : "kitten.", "n" : "cat" }
          ]
        }
      ]
    }
        expected_plain_table = """\
+---+---+-------+---------+
| A | A | black | cat     |
| B | A | white | kitten. |
+---+---+-------+---------+"""
        plain_table = str(collate_pretokenized_json(json_in, output="table"))
        self.assertEqual(expected_plain_table, plain_table)

    def testHTMLOutputVerticalLayoutPretokenizedJSON(self):
        json_in = {
      "witnesses" : [
        {
          "id" : "A",
          "tokens" : [
              { "t" : "A", "ref" : 123 },
              { "t" : "black" , "adj" : True },
              { "t" : "cat", "id" : "xyz" }
          ]
        },
        {
          "id" : "B",
          "tokens" : [
              { "t" : "A" },
              { "t" : "white" , "adj" : True },
              { "t" : "kitten.", "n" : "cat" }
          ]
        }
      ]
    }
        expected_output = """\
+-------+---------+
|   A   |    B    |
+-------+---------+
|   A   |    A    |
+-------+---------+
| black |  white  |
+-------+---------+
|  cat  | kitten. |
+-------+---------+"""
        plain_text_output = str(collate_pretokenized_json(json_in, layout="vertical"))
        self.assertEquals(expected_output, plain_text_output)

    def testSegmentationPretokenizedJSON(self):
        json_in = {
      "witnesses" : [
        {
          "id" : "A",
          "tokens" : [
              { "t" : "A", "ref" : 123 },
              { "t" : "black" , "adj" : True },
              { "t" : "cat", "id" : "xyz" }
          ]
        },
        {
          "id" : "B",
          "tokens" : [
              { "t" : "A" },
              { "t" : "white" , "adj" : True },
              { "t" : "stripy", "adj" : True },
              { "t" : "kitten.", "n" : "cat" }
          ]
        }
      ]
    }
        bad_func = lambda x: collate_pretokenized_json(x, segmentation=True)
        self.assertRaises(UnsupportedError, bad_func, json_in )

if __name__ == '__main__':
    unittest.main()



'''
Created on Jun 11, 2014

@author: Ronald Haentjens Dekker
'''
import unittest
from collatex import Collation


class Test(unittest.TestCase):

    def testCachingOfSuffixArrayAndLCPArray(self):
        collation = Collation()
        collation.add_plain_witness("A", "content")
        collation.add_plain_witness("B", "content")
        sa1 = collation.get_sa()
        sa2 = collation.get_sa()
        self.assertEquals(sa1,sa2)
        collation.add_plain_witness("C", "content")
        sa3 = collation.get_sa()
        self.assertNotEquals(sa2, sa3)


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
Created on Apr 7, 2014

@author: Ronald Haentjens Dekker
'''
import unittest
from collatex.linsuffarr import SuffixArray



class Test(unittest.TestCase):
    pass

    #TODO: write asserts
    def testLinearSuffixArray(self):
        sa = SuffixArray("The Quick Brown Fox Jumped Over")
        print(sa)
        pass

    #TODO: write asserts
    def testSA2(self):
        # $ is meant to separate witnesses here
        sa = SuffixArray("a b c d F g h i ! K ! q r s t $ a b c d F g h i ! q r s t")
        print(sa)
        print(sa._LCP_values)
        
if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


#
# Copyright CEA/DAM/DIF (2012)
#  Contributor: Stephane THIELL <stephane.thiell@cea.fr>
#
# This file is part of the ClusterShell library.
#
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-C license and that you accept its terms.

# This class has been modified to make it Python 3 compliant by Ronald Haentjens Dekker

"""
Cluster range set module.

Instances of RangeSet provide similar operations than the builtin set type,
extended to support cluster ranges-like format and stepping support ("0-8/2").
"""

__all__ = ['RangeSetException',
           'RangeSetParseError',
           'RangeSetPaddingError',
           'RangeSet']


class RangeSetException(Exception):
    """Base RangeSet exception class."""

class RangeSetParseError(RangeSetException):
    """Raised when RangeSet parsing cannot be done properly."""
    def __init__(self, part, msg):
        if part:
            msg = "%s : \"%s\"" % (msg, part)
        RangeSetException.__init__(self, msg)
        # faulty subrange; this allows you to target the error
        self.part = part

class RangeSetPaddingError(RangeSetParseError):
    """Raised when a fatal padding incoherency occurs"""
    def __init__(self, part, msg):
        RangeSetParseError.__init__(self, part, "padding mismatch (%s)" % msg)


class RangeSet(set):
    """
    Mutable set of cluster node indexes featuring a fast range-based API.
    
    This class aims to ease the management of potentially large cluster range
    sets and is used by the NodeSet class.

    RangeSet basic constructors:
       >>> rset = RangeSet()            # empty RangeSet
       >>> rset = RangeSet("5,10-42")   # contains 5, 10 to 42
       >>> rset = RangeSet("0-10/2")    # contains 0, 2, 4, 6, 8, 10

    Since v1.6, any iterable of integers can be specified as first argument:
       >>> RangeSet([3, 6, 8, 7, 1])
       1,3,6-8
       >>> rset2 = RangeSet(rset)

    Padding of ranges (eg. "003-009") can be managed through a public RangeSet
    instance variable named padding. It may be changed at any time. Since v1.6,
    padding is a simple display feature per RangeSet object, thus current
    padding value is not taken into account when computing set operations.
    Since v1.6, RangeSet is itself an iterator over its items as integers
    (instead of strings). To iterate over string items as before (with
    optional padding), you can now use the RangeSet.striter() method.

    RangeSet provides methods like union(), intersection(), difference(),
    symmetric_difference() and their in-place versions update(),
    intersection_update(), difference_update(),
    symmetric_difference_update() which conform to the Python Set API.
    """
    _VERSION = 3    # serial version number

    # define __new__() to workaround built-in set subclassing with Python 2.4
    def __new__(cls, pattern=None, autostep=None):
        """Object constructor"""
        return set.__new__(cls)
        
    def __init__(self, pattern=None, autostep=None):
        """Initialize RangeSet with optional string pattern and autostep
        threshold.
        """
        if pattern is None or isinstance(pattern, str):
            set.__init__(self)
        else:
            set.__init__(self, pattern)

        if isinstance(pattern, RangeSet):
            self._autostep = pattern._autostep
            self.padding = pattern.padding
        else:
            self._autostep = None
            self.padding = None
        self.autostep = autostep

        if isinstance(pattern, str):
            self._parse(pattern)

    def _parse(self, pattern):
        """Parse string of comma-separated x-y/step -like ranges"""
        # Comma separated ranges
        if pattern.find(',') < 0:
            subranges = [pattern]
        else:
            subranges = pattern.split(',')

        for subrange in subranges:
            if subrange.find('/') < 0:
                step = 1
                baserange = subrange
            else:
                baserange, step = subrange.split('/', 1)

            try:
                step = int(step)
            except ValueError:
                raise RangeSetParseError(subrange,
                        "cannot convert string to integer")

            if baserange.find('-') < 0:
                if step != 1:
                    raise RangeSetParseError(subrange,
                            "invalid step usage")
                begin = end = baserange
            else:
                begin, end = baserange.split('-', 1)

            # compute padding and return node range info tuple
            try:
                pad = 0
                if int(begin) != 0:
                    begins = begin.lstrip("0")
                    if len(begin) - len(begins) > 0:
                        pad = len(begin)
                    start = int(begins)
                else:
                    if len(begin) > 1:
                        pad = len(begin)
                    start = 0
                if int(end) != 0:
                    ends = end.lstrip("0")
                else:
                    ends = end
                stop = int(ends)
            except ValueError:
                raise RangeSetParseError(subrange,
                        "cannot convert string to integer")

            # check preconditions
            if stop > 1e100 or start > stop or step < 1:
                raise RangeSetParseError(subrange,
                                         "invalid values in range")

            self.add_range(start, stop + 1, step, pad)
        
    @classmethod
    def fromlist(cls, rnglist, autostep=None):
        """Class method that returns a new RangeSet with ranges from provided
        list."""
        inst = RangeSet(autostep=autostep)
        inst.updaten(rnglist)
        return inst

    @classmethod
    def fromone(cls, index, pad=0, autostep=None):
        """Class method that returns a new RangeSet of one single item or
        a single range (from integer or slice object)."""
        inst = RangeSet(autostep=autostep)
        # support slice object with duck-typing
        try:
            inst.add(index, pad)
        except TypeError:
            if not index.stop:
                raise ValueError("Invalid range upper limit (%s)" % index.stop)
            inst.add_range(index.start or 0, index.stop, index.step or 1, pad)
        return inst

    def get_autostep(self):
        """Get autostep value (property)"""
        if self._autostep >= 1E100:
            return None
        else:
            return self._autostep + 1

    def set_autostep(self, val):
        """Set autostep value (property)"""
        if val is None:
            # disabled by default for pdsh compat (+inf is 1E400, but a bug in
            # python 2.4 makes it impossible to be pickled, so we use less)
            # NOTE: Later, we could consider sys.maxint here
            self._autostep = 1E100
        else:
            # - 1 because user means node count, but we means real steps
            self._autostep = int(val) - 1

    autostep = property(get_autostep, set_autostep)
    
    def _sorted(self):
        """Get sorted list from inner set."""
        return sorted(set.__iter__(self))

    def __iter__(self):
        """Iterate over each element in RangeSet."""
        return iter(self._sorted())

    def striter(self):
        """Iterate over each (optionally padded) string element in RangeSet."""
        pad = self.padding or 0
        for i in self._sorted():
            yield "%0*d" % (pad, i)

    def contiguous(self):
        """Object-based iterator over contiguous range sets."""
        pad = self.padding or 0
        for sli in self._contiguous_slices():
            yield RangeSet.fromone(slice(sli.start, sli.stop, sli.step), pad)

    def __reduce__(self):
        """Return state information for pickling."""
        return self.__class__, (str(self),), \
            { 'padding': self.padding, \
              '_autostep': self._autostep, \
              '_version' : RangeSet._VERSION }

    def __setstate__(self, dic):
        """called upon unpickling"""
        self.__dict__.update(dic)
        if getattr(self, '_version', 0) < RangeSet._VERSION:
            # unpickle from old version?
            if getattr(self, '_version', 0) <= 1:
                # v1 (no object versioning) - CSv1.3
                setattr(self, '_ranges', [(slice(start, stop + 1, step), pad) \
                    for start, stop, step, pad in getattr(self, '_ranges')])
            elif hasattr(self, '_ranges'):
                # v2 - CSv1.4-1.5
                self_ranges = getattr(self, '_ranges')
                if self_ranges and type(self_ranges[0][0]) is not slice:
                    # workaround for object pickled from Python < 2.5
                    setattr(self, '_ranges', [(slice(start, stop, step), pad) \
                        for (start, stop, step), pad in self_ranges])
            # convert to v3
            for sli, pad in getattr(self, '_ranges'):
                self.add_range(sli.start, sli.stop, sli.step, pad)
            delattr(self, '_ranges')
            delattr(self, '_length')

    def _strslices(self):
        """Stringify slices list (x-y/step format)"""
        pad = self.padding or 0
        for sli in self.slices():
            if sli.start + 1 == sli.stop:
                yield "%0*d" % (pad, sli.start)
            else:
                assert sli.step >= 0, "Internal error: sli.step < 0"
                if sli.step == 1:
                    yield "%0*d-%0*d" % (pad, sli.start, pad, sli.stop - 1)
                else:
                    yield "%0*d-%0*d/%d" % (pad, sli.start, pad, sli.stop - 1, \
                                            sli.step)
        
    def __str__(self):
        """Get comma-separated range-based string (x-y/step format)."""
        return ','.join(self._strslices())

    # __repr__ is the same as __str__ as it is a valid expression that
    # could be used to recreate a RangeSet with the same value
    __repr__ = __str__

    def _contiguous_slices(self):
        """Internal iterator over contiguous slices in RangeSet."""
        k = j = None
        for i in self._sorted():
            if k is None:
                k = j = i
            if i - j > 1:
                yield slice(k, j + 1, 1)
                k = i
            j = i
        if k is not None:
            yield slice(k, j + 1, 1)

    def _folded_slices(self):
        """Internal generator that is able to retrieve ranges organized by step.
        Complexity: O(n) with n = number of ranges in tree."""
        if len(self) == 0:
            return

        prng = None         # pending range
        istart = None       # processing starting indice
        m = 0               # processing step
        for sli in self._contiguous_slices():
            start = sli.start
            stop = sli.stop
            unitary = (start + 1 == stop)   # one indice?
            if istart is None:  # first loop
                if unitary:
                    istart = start
                else:
                    prng = [start, stop, 1]
                    istart = stop - 1
                i = k = istart
            elif m == 0:        # istart is set but step is unknown
                if not unitary:
                    if prng is not None:
                        # yield and replace pending range
                        yield slice(*prng)
                    else:
                        yield slice(istart, istart + 1, 1)
                    prng = [start, stop, 1]
                    istart = k = stop - 1
                    continue
                i = start
            else:               # step m > 0
                assert m > 0
                i = start
                # does current range lead to broken step?
                if m != i - k or not unitary:
                    #j = i if m == i - k else k
                    if m == i - k: j = i
                    else: j = k
                    # stepped is True when autostep setting does apply
                    stepped = (j - istart >= self._autostep * m)
                    if prng:    # yield pending range?
                        if stepped:
                            prng[1] -= 1
                        else:
                            istart += m
                        yield slice(*prng)
                        prng = None
                if m != i - k:
                    # case: step value has changed
                    if stepped:
                        yield slice(istart, k + 1, m)
                    else:
                        for j in range(istart, k - m + 1, m):
                            yield slice(j, j + 1, 1)
                        if not unitary:
                            yield slice(k, k + 1, 1)
                    if unitary:
                        if stepped:
                            istart = i = k = start
                        else:
                            istart = k
                    else:
                        prng = [start, stop, 1]
                        istart = i = k = stop - 1
                elif not unitary:
                    # case: broken step by contiguous range
                    if stepped:
                        # yield 'range/m' by taking first indice of new range
                        yield slice(istart, i + 1, m)
                        i += 1
                    else:
                        # autostep setting does not apply in that case
                        for j in range(istart, i - m + 1, m):
                            yield slice(j, j + 1, 1)
                    if stop > i + 1:    # current->pending only if not unitary
                        prng = [i, stop, 1]
                    istart = i = k = stop - 1
            m = i - k   # compute step
            k = i
        # exited loop, process pending range or indice...
        if m == 0:
            if prng:
                yield slice(*prng)
            else:
                yield slice(istart, istart + 1, 1)
        else:
            assert m > 0
            stepped = (k - istart >= self._autostep * m)
            if prng:
                if stepped:
                    prng[1] -= 1
                else:
                    istart += m
                yield slice(*prng)
                prng = None
            if stepped:
                yield slice(istart, i + 1, m)
            else:
                for j in range(istart, i + 1, m):
                    yield slice(j, j + 1, 1)

    def slices(self):
        """
        Iterate over RangeSet ranges as Python slice objects.
        """
        # return an iterator
        if self._autostep >= 1E100:
            return self._contiguous_slices()
        else:
            return self._folded_slices()

    def __getitem__(self, index):
        """
        Return the element at index or a subrange when a slice is specified.
        """
        if isinstance(index, slice):
            inst = RangeSet()
            inst._autostep = self._autostep
            inst.padding = self.padding
            inst.update(self._sorted()[index])
            return inst
        elif isinstance(index, int):
            return self._sorted()[index]
        else:
            raise TypeError("%s indices must be integers" % self.__class__.__name__)

    def split(self, nbr):
        """
        Split the rangeset into nbr sub-rangesets (at most). Each
        sub-rangeset will have the same number of elements more or
        less 1. Current rangeset remains unmodified. Returns an
        iterator.

        >>> RangeSet("1-5").split(3) 
        RangeSet("1-2")
        RangeSet("3-4")
        RangeSet("foo5")
        """
        assert(nbr > 0)

        # We put the same number of element in each sub-nodeset.
        slice_size = len(self) / nbr
        left = len(self) % nbr

        begin = 0
        for i in range(0, min(nbr, len(self))):
            length = slice_size + int(i < left)
            yield self[begin:begin + length]
            begin += length

    def add_range(self, start, stop, step=1, pad=0):
        """
        Add a range (start, stop, step and padding length) to RangeSet.
        Like the Python built-in function range(), the last element is
        the largest start + i * step less than stop.
        """
        assert start < stop, "please provide ordered node index ranges"
        assert step > 0
        assert pad >= 0
        assert stop - start < 1e9, "range too large"

        if pad > 0 and self.padding is None:
            self.padding = pad
        set.update(self, range(start, stop, step))

    def copy(self):
        """Return a shallow copy of a RangeSet."""
        cpy = self.__class__()
        cpy._autostep = self._autostep
        cpy.padding = self.padding
        cpy.update(self)
        return cpy

    __copy__ = copy # For the copy module

    def __eq__(self, other):
        """
        RangeSet equality comparison.
        """
        # Return NotImplemented instead of raising TypeError, to
        # indicate that the comparison is not implemented with respect
        # to the other type (the other comparand then gets a change to
        # determine the result, then it falls back to object address
        # comparison).
        if not isinstance(other, RangeSet):
            return NotImplemented
        return len(self) == len(other) and self.issubset(other)

    # Standard set operations: union, intersection, both differences.
    # Each has an operator version (e.g. __or__, invoked with |) and a
    # method version (e.g. union).
    # Subtle:  Each pair requires distinct code so that the outcome is
    # correct when the type of other isn't suitable.  For example, if
    # we did "union = __or__" instead, then Set().union(3) would return
    # NotImplemented instead of raising TypeError (albeit that *why* it
    # raises TypeError as-is is also a bit subtle).

    def _wrap_set_op(self, fun, arg):
        """Wrap built-in set operations for RangeSet to workaround built-in set
        base class issues (RangeSet.__new/init__ not called)"""
        result = fun(self, arg)
        result._autostep = self._autostep
        result.padding = self.padding
        return result

    def __or__(self, other):
        """Return the union of two RangeSets as a new RangeSet.

        (I.e. all elements that are in either set.)
        """
        if not isinstance(other, set):
            return NotImplemented
        return self.union(other)

    def union(self, other):
        """Return the union of two RangeSets as a new RangeSet.

        (I.e. all elements that are in either set.)
        """
        return self._wrap_set_op(set.union, other)

    def __and__(self, other):
        """Return the intersection of two RangeSets as a new RangeSet.

        (I.e. all elements that are in both sets.)
        """
        if not isinstance(other, set):
            return NotImplemented
        return self.intersection(other)

    def intersection(self, other):
        """Return the intersection of two RangeSets as a new RangeSet.

        (I.e. all elements that are in both sets.)
        """
        #NOTE: This is a work around 
        # Python 3 return as the result of set.intersection a new set instance.
        # Python 2 however returns as a the result a ClusterShell.RangeSet.RangeSet instance.
        # ORIGINAL CODE: return self._wrap_set_op(set.intersection, other)
        copy = self.copy()
        copy.intersection_update(other)
        return copy

    def __xor__(self, other):
        """Return the symmetric difference of two RangeSets as a new RangeSet.

        (I.e. all elements that are in exactly one of the sets.)
        """
        if not isinstance(other, set):
            return NotImplemented
        return self.symmetric_difference(other)

    def symmetric_difference(self, other):
        """Return the symmetric difference of two RangeSets as a new RangeSet.
        
        (ie. all elements that are in exactly one of the sets.)
        """
        return self._wrap_set_op(set.symmetric_difference, other)

    def  __sub__(self, other):
        """Return the difference of two RangeSets as a new RangeSet.

        (I.e. all elements that are in this set and not in the other.)
        """
        if not isinstance(other, set):
            return NotImplemented
        return self.difference(other)

    def difference(self, other):
        """Return the difference of two RangeSets as a new RangeSet.

        (I.e. all elements that are in this set and not in the other.)
        """
        return self._wrap_set_op(set.difference, other)

    # Membership test

    def __contains__(self, element):
        """Report whether an element is a member of a RangeSet.
        Element can be either another RangeSet object, a string or an
        integer.

        (Called in response to the expression `element in self'.)
        """
        if isinstance(element, set):
            return element.issubset(self)

        return set.__contains__(self, int(element))

    # Subset and superset test

    def issubset(self, other):
        """Report whether another set contains this RangeSet."""
        self._binary_sanity_check(other)
        return set.issubset(self, other)

    def issuperset(self, other):
        """Report whether this RangeSet contains another set."""
        self._binary_sanity_check(other)
        return set.issuperset(self, other)

    # Inequality comparisons using the is-subset relation.
    __le__ = issubset
    __ge__ = issuperset

    def __lt__(self, other):
        self._binary_sanity_check(other)
        return len(self) < len(other) and self.issubset(other)

    def __gt__(self, other):
        self._binary_sanity_check(other)
        return len(self) > len(other) and self.issuperset(other)

    # Assorted helpers

    def _binary_sanity_check(self, other):
        """Check that the other argument to a binary operation is also  a set,
        raising a TypeError otherwise."""
        if not isinstance(other, set):
            raise TypeError("Binary operation only permitted between sets")

    # In-place union, intersection, differences.
    # Subtle:  The xyz_update() functions deliberately return None,
    # as do all mutating operations on built-in container types.
    # The __xyz__ spellings have to return self, though.
    
    def __ior__(self, other):
        """Update a RangeSet with the union of itself and another."""
        self._binary_sanity_check(other)
        set.__ior__(self, other)
        return self

    def union_update(self, other):
        """Update a RangeSet with the union of itself and another."""
        self.update(other)

    def __iand__(self, other):
        """Update a RangeSet with the intersection of itself and another."""
        self._binary_sanity_check(other)
        set.__iand__(self, other)
        return self

    def intersection_update(self, other):
        """Update a RangeSet with the intersection of itself and another."""
        set.intersection_update(self, other)

    def __ixor__(self, other):
        """Update a RangeSet with the symmetric difference of itself and
        another."""
        self._binary_sanity_check(other)
        set.symmetric_difference_update(self, other)
        return self

    def symmetric_difference_update(self, other):
        """Update a RangeSet with the symmetric difference of itself and
        another."""
        set.symmetric_difference_update(self, other)
        
    def __isub__(self, other):
        """Remove all elements of another set from this RangeSet."""
        self._binary_sanity_check(other)
        set.difference_update(self, other)
        return self

    def difference_update(self, other, strict=False):
        """Remove all elements of another set from this RangeSet.
        
        If strict is True, raise KeyError if an element cannot be removed.
        (strict is a RangeSet addition)"""
        if strict and other not in self:
            raise KeyError(other.difference(self)[0])
        set.difference_update(self, other)

    # Python dict-like mass mutations: update, clear

    def update(self, iterable):
        """Add all integers from an iterable (such as a list)."""
        if isinstance(iterable, RangeSet):
            # keep padding unless is has not been defined yet
            if self.padding is None and iterable.padding is not None:
                self.padding = iterable.padding
        assert type(iterable) is not str
        set.update(self, iterable)

    def updaten(self, rangesets):
        """
        Update a rangeset with the union of itself and several others.
        """
        for rng in rangesets:
            if isinstance(rng, set):
                self.update(rng)
            else:
                self.update(RangeSet(rng))
            # py2.5+
            #self.update(rng if isinstance(rng, set) else RangeSet(rng))

    def clear(self):
        """Remove all elements from this RangeSet."""
        set.clear(self)
        self.padding = None

    # Single-element mutations: add, remove, discard

    def add(self, element, pad=0):
        """Add an element to a RangeSet.
        This has no effect if the element is already present.
        """
        set.add(self, int(element))
        if pad > 0 and self.padding is None:
            self.padding = pad

    def remove(self, element):
        """Remove an element from a RangeSet; it must be a member.
        
        Raise KeyError if element is not contained in RangeSet.
        Raise ValueError if element is not castable to integer.
        """
        set.remove(self, int(element))

    def discard(self, element):
        """Remove element from the RangeSet if it is a member.

        If the element is not a member, do nothing.
        """
        try:
            i = int(element)
            set.discard(self, i)
        except ValueError:
            pass # ignore other object types




'''
Created on Aug 5, 2014

@author: Ronald Haentjens Dekker
'''
from collatex.collatex_suffix import Occurrence, BlockWitness, Block,\
    PartialOverlapException
from operator import attrgetter
from ClusterShell.RangeSet import RangeSet
# optionally load the Levenshtein dependency for near match functionality.
# Consists of C code, needs to be compiled which is problematic on Windows.
# There are pre-compiled binaries however, but this requires an extra step
# during installation.
try:
    from Levenshtein import ratio
except:
    pass


'''
 This class scores cells in the edit graph table based on the largest non overlapping blocks
 that are found in the witness set.
'''

class Scorer(object):
    def __init__(self, collation, near_match=False):
        self.collation = collation
        self.blocks = []
        self.global_tokens_to_occurrences = {}
        if near_match:
            self.match_function = self.near_match
        else:
            self.match_function = self.match
        
    def prepare_witness(self, witness):
        # this code can probably done more efficient, for now the main goal is to make it work
        block_witness = self._get_block_witness(witness)
        tokens_to_occurrences = self._build_tokens_to_occurrences(self.collation, witness, block_witness)
        #NOTE: we do not have to store all tokens from the witness
        #NOTE: if we split the dict into two: one for next witness and one for superbase
        #NOTE: only the ones that are going to be aligned have to be stored in superbase dict.
        self.global_tokens_to_occurrences.update(tokens_to_occurrences)
        
    # return values:
    # 0 = FULL_MATCH
    # -1 = NO MATCH
    # 1 = PARTIAL MATCH
    def match(self, token_a, token_b):
        # now we need to determine whether this node represents a match
        # determine this based on whether token a and token b are part of the same block
        occur_a = self.global_tokens_to_occurrences.setdefault(token_a, None)
        occur_b = self.global_tokens_to_occurrences.setdefault(token_b, None)
        if occur_a and occur_b:
            match = occur_a.block == occur_b.block
        else:
            match = False
        if match:
            return 0
        else:
            return -1

    def near_match(self, token_a, token_b):
        result = self.match(token_a, token_b)
        if result==0:
            return 0
        r = ratio(token_a.token_string, token_b.token_string)
        print(str(token_a)+" "+str(token_b)+" "+str(r))
        if r > 0.6: 
            return 1
        else:
            return -1
        pass
    
    # edit operation: 
    #    0 == match/replacement
    #    1 == addition/omission
    def score_cell(self, table_node, parent_node, token_a, token_b, y, x, edit_operation):
        # no matching possible in this case (always treated as a gap)
        # it is either an add or a delete
        if x == 0 or y == 0:
            table_node.g = parent_node.g - 1
            return
        
        # it is either an add/delete or replacement (so an add and a delete)
        # it is a replacement
        if edit_operation == 0:
            match = self.match_function(token_a, token_b)
#             print("testing "+token_a.token_string+" and "+token_b.token_string+" "+str(match))   
            # match = token_a.token_string == token_b.token_string
            # based on match or not and parent_node calculate new score
            if match==0:
                # mark the fact that this node is match
                table_node.match = True
                # do not change score for now 
                table_node.g = parent_node.g
                # count segments
                if parent_node.match == False:
                    table_node.segments = parent_node.segments + 1
                return
            if match==1:
                table_node.g = parent_node.g - 0.5 #TODO: TEST TEST TEST
                pass
            else:
                table_node.g = parent_node.g - 2
                return
        # it is an add/delete
        else:
            table_node.g = parent_node.g - 1
            return 

    #TODO: it should be possible to do this simpler, faster
    # An occurrence should know its tokens, since it knows its token range
    def _build_tokens_to_occurrences(self, collation, witness, block_witness):
        tokens_to_occurrence = {}
        witness_range = collation.get_range_for_witness(witness.sigil)
        token_counter = witness_range[0]
        # note: this can be done faster by focusing on the occurrences
        # instead of the tokens
        for token in witness.tokens():
            for occurrence in block_witness.occurrences:
                if occurrence.is_in_range(token_counter):
                    tokens_to_occurrence[token]=occurrence
            token_counter += 1
        return tokens_to_occurrence

    

    '''
    Internal method to transform a Witness into a Block Witness.
    '''
    def _get_block_witness(self, witness):
#         print("Generating block witness for: "+witness.sigil)
        sigil_witness = witness.sigil
        range_witness = self.collation.get_range_for_witness(sigil_witness)
        #NOTE: to prevent recalculation of blocks
        if not self.blocks:
            self.blocks = self._get_non_overlapping_repeating_blocks() 
        blocks = self.blocks 
        # make a selection of blocks and occurrences of these blocks in the selected witness
        occurrences = []
        for block in blocks:
            block_ranges_in_witness = block.ranges & range_witness
            # note this are multiple ranges
            # we need to iterate over every single one
            for block_range in block_ranges_in_witness.contiguous():
                occurrence = Occurrence(block_range, block)
                occurrences.append(occurrence) 
        # sort occurrences on position
        sorted_o = sorted(occurrences, key=attrgetter('lower_end'))
        block_witness = BlockWitness(sorted_o, self.collation.tokens)
        return block_witness

    '''
    Find all the non overlapping repeating blocks in the witness set of this collation.
    '''
    def _get_non_overlapping_repeating_blocks(self):
        extended_suffix_array = self.collation.to_extended_suffix_array()
        potential_blocks = extended_suffix_array.split_lcp_array_into_intervals() 
        self.filter_potential_blocks(potential_blocks)
        # step 3: sort the blocks based on depth (number of repetitions) first,
        # second length of LCP interval,
        # third sort on parent LCP interval occurrences.
        sorted_blocks_on_priority = sorted(potential_blocks, key=attrgetter("number_of_occurrences", "minimum_block_length", "number_of_siblings"), reverse=True)
        # step 4: select the definitive blocks
        occupied = RangeSet()
        real_blocks = []
        for potential_block in sorted_blocks_on_priority:
#           print(potential_block.info())
            try:
                non_overlapping_range = potential_block.calculate_non_overlapping_range_with(occupied)
                if non_overlapping_range:
#                     print("Selecting: "+str(potential_block))
                    occupied.union_update(non_overlapping_range)
                    real_blocks.append(Block(non_overlapping_range))
            except PartialOverlapException:          
#                 print("Skip due to conflict: "+str(potential_block))
                while potential_block.minimum_block_length > 1:
                    # retry with a different length: one less
                    for idx in range(potential_block.start+1, potential_block.end+1):
                        potential_block.LCP[idx] -= 1
                    potential_block.length -= 1
                    try:
                        non_overlapping_range = potential_block.calculate_non_overlapping_range_with(occupied)
                        if non_overlapping_range:
#                             print("Retried and selecting: "+str(potential_block))
                            occupied.union_update(non_overlapping_range)
                            real_blocks.append(Block(non_overlapping_range))
                            break
                    except PartialOverlapException:          
#                         print("Retried and failed again")
                        pass
        return real_blocks


    # filter out all the blocks that have more than one occurrence within a witness
    def filter_potential_blocks(self, potential_blocks):
        for potential_block in potential_blocks[:]:
            for witness in self.collation.witnesses:
                witness_sigil = witness.sigil
                witness_range = self.collation.get_range_for_witness(witness_sigil)
                inter = witness_range.intersection(potential_block.block_occurrences())
                if potential_block.number_of_occurrences > len(self.collation.witnesses) or len(inter)> potential_block.minimum_block_length:
#                     print("Removing block: "+str(potential_block))
                    potential_blocks.remove(potential_block)
                    break




'''
Created on Sep 12, 2014

@author: Ronald Haentjens Dekker
'''
import unittest
from collatex import Collation, collate


class Test(unittest.TestCase):

    def test_exact_matching(self):
        collation = Collation()
        collation.add_plain_witness("A", "I bought this glass, because it matches those dinner plates")
        collation.add_plain_witness("B", "I bought those glasses")
        alignment_table = collate(collation)
        self.assertEquals(["I bought", "this glass, because it matches", "those", "dinner plates"],
                          alignment_table.rows[0].to_list())
        self.assertEquals(["I bought", "-", "those", "glasses"], alignment_table.rows[1].to_list())

    def test_near_matching(self):
        collation = Collation()
        collation.add_plain_witness("A", "I bought this glass, because it matches those dinner plates")
        collation.add_plain_witness("B", "I bought those glasses")
        alignment_table = collate(collation, near_match=True)
        self.assertEquals(["I bought", "this glass, because it matches those dinner plates"],
                          alignment_table.rows[0].to_list())
        self.assertEquals(["I bought", "those glasses"], alignment_table.rows[1].to_list())


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testOmission']
    unittest.main()



'''
Created on Aug 4, 2014

@author: Ronald
'''
import unittest
from tests import unit_disabled
from collatex.core_functions import VariantGraph
from collatex.edit_graph_aligner import EditGraphAligner
from collatex import Collation


class Test(unittest.TestCase):

    # global score
    def assertRow(self, expected, cell_data):
        actual = []
        for cell in cell_data:
            actual.append(cell.g)
        self.assertEqual(expected, actual)
    
    def assertSuperbaseEquals(self, expected, superbase):
        actual = ""
        for token in superbase:
            if actual:
                actual += " "
            actual += str(token)
        self.assertEquals(expected, actual)    
    
    def debugRowSegments(self, cell_data):
        actual = []
        for cell in cell_data:
            actual.append(cell.segments)
        print(actual)
    
    def debug_table(self, aligner, table):
        for y in range(aligner.length_witness_b+1):
            for x in range(aligner.length_witness_a+1):
                print (y, x), table[y][x]


    # we need to introduce a gap here
    def testOmission(self):
        collation = Collation()
        collation.add_plain_witness("A", "a b c")
        collation.add_plain_witness("B", "b c")
        aligner = EditGraphAligner(collation)
        graph = VariantGraph()
        aligner.collate(graph, collation)
        table = aligner.table
#         self.debug_table(aligner, table)
        self.assertEqual(0, table[0][0].g)
        self.assertEqual(-1, table[0][1].g)
        self.assertEqual(-2, table[0][2].g)
        self.assertEqual(-3, table[0][3].g)
        self.assertEqual(-1, table[1][0].g)
        self.assertEqual(-2, table[1][1].g)
        self.assertEqual(-1, table[1][2].g)
        self.assertEqual(-2, table[1][3].g)
        self.assertEqual(-2, table[2][0].g)
        self.assertEqual(-3, table[2][1].g)
        self.assertEqual(-2, table[2][2].g)
        self.assertEqual(-1, table[2][3].g)


    # note: the scoring table in this test is only correct when block detection is OFF
    @unit_disabled
    def testOmission2GlobalScore(self):
        collation = Collation()
        collation.add_plain_witness("A", "a a b c")
        collation.add_plain_witness("B", "a b c")
        aligner = EditGraphAligner(collation)
        graph = VariantGraph()
        aligner.collate(graph, collation)
        table = aligner.table
 
        self.assertRow([0, -1, -2, -3, -4], table[0])
        self.assertRow([-1, 0, -1, -2, -3], table[1])
        self.assertRow([-2, -1, -2, -1, -2], table[2])
        self.assertRow([-3, -2, -3, -2, -1], table[3])


    def test_superbase(self):
        collation = Collation()
        collation.add_plain_witness("A", "X a b c d e f X g h i Y Z j k")
        collation.add_plain_witness("B", "a b c Y d e f Y Z g h i X j k")
        aligner = EditGraphAligner(collation)
        graph = VariantGraph()
        aligner.collate(graph, collation)
        superbase = aligner.new_superbase
        self.assertSuperbaseEquals("X a b c Y d e f X Y Z g h i Y Z X j k", superbase)
        
        
        # TODO: add Y to the witness B (to check end modification
        
        
#     def test_path(self):
#         a = Witness("A", "a b c")
#         b = Witness("B", "a b c")
#         aligner = EditGraphAligner(a, b)
#         aligner.align()
#         segments = aligner.get_segments()
#         self.assertSegments(["a b c"], segments)
        
        
#         path = aligner.get_path()
#         self.assertPath([(0,0),(1,1),(2,2),(3,3)], path)
        
#     def testOmission2SegmentsScore(self):
#         a = Witness("A", "a a b c")
#         b = Witness("B", "a b c")
#         aligner = EditGraphAligner(a, b)
#         aligner.align()
#         table = aligner.table
#
#         self.debugRowSegments(table[0])
#         self.debugRowSegments(table[1])
            
#      TODO: add test for segments
#         self.assertEqual(1, table[3][4].segments)
        
        
        
if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testOmission']
    unittest.main()



'''
Created on May 3, 2014

@author: Ronald Haentjens Dekker
'''
import unittest
from networkx.drawing.nx_pydot import to_pydot
from collatex.core_classes import VariantGraph
from collatex import Collation
#from collatex.dekker_suffix_algorithm import DekkerSuffixAlgorithm


class Test(unittest.TestCase):

    #TODO: this test is not finished! 
    # def test_Hermans_case_variantgraph(self):
    #     collation = Collation()
    #     collation.add_plain_witness("W1", "a b c d F g h i ! K ! q r s t")
    #     collation.add_plain_witness("W2", "a b c d F g h i ! q r s t")
    #     graph = VariantGraph()
    #     algorithm = DekkerSuffixAlgorithm(collation)
    #     algorithm.build_variant_graph_from_blocks(graph, collation)
    #     start_vertex = graph.start
    #     a = graph.vertexWith("a")
    #     b = graph.vertexWith("b")
    #     t = graph.vertexWith("t")
    #     end_vertex = graph.end
    #     self.assert_(graph.edge_between(start_vertex, a))
    #     self.assert_(graph.edge_between(a, b))
    #     self.assert_(graph.edge_between(t, end_vertex))
    #     #print(len(graph.vertices()))
    #     #print(len(graph.edges()))
    #
    #     # display graph
    #     #print(graph.graph.nodes())
    #     #view_pygraphviz(graph.graph)
    #
#         dot = to_pydot(graph.graph)
#         dot.write("rawoutput")

#     def test_variant_graph_two_equal_witnesses(self):
#         collation = Collation()
#         collation.add_plain_witness('A', 'the black cat')
#         collation.add_plain_witness('B', 'the black cat')
#         graph = collation.collate()
#         the_vertex = vertexWith(graph, 'the')
#         black_vertex = vertexWith(graph, 'black')
#         cat_vertex = vertexWith(graph, 'cat')
#         
#         self.assert_(graph.edge_between(graph.start, the_vertex))
#         self.assert_(graph.edge_between(the_vertex, black_vertex))
#         self.assert_(graph.edge_between(cat_vertex, graph.end))
#       
#         @Test
#   public void twoWitnesses() {
#     final SimpleWitness[] w = createWitnesses("the black cat", "the black cat");
#     final VariantGraph graph = collate(w);
# 
#     assertEquals(5, Iterables.size(graph.vertices()));
#     assertEquals(4, Iterables.size(graph.edges()));
# 
#     final VariantGraph.Vertex theVertex = vertexWith(graph, "the", w[0]);
#     final VariantGraph.Vertex blackVertex = vertexWith(graph, "black", w[0]);
#     final VariantGraph.Vertex catVertex = vertexWith(graph, "cat", w[0]);
# 
#     assertHasWitnesses(edgeBetween(graph.getStart(), theVertex), w[0], w[1]);
#     assertHasWitnesses(edgeBetween(theVertex, blackVertex), w[0], w[1]);
#     assertHasWitnesses(edgeBetween(blackVertex, catVertex), w[0], w[1]);
#     assertHasWitnesses(edgeBetween(catVertex, graph.getEnd()), w[0], w[1]);
    pass

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
Created on Aug 4, 2014

@author: Ronald Haentjens Dekker
'''

from collatex.astar import AStarNode, AStar

#The second parameter could also be the tree
class DecisionTreeNode(AStarNode):
    def __init__(self, aligner):
        self.pointer_a = 0
        self.pointer_b = 0
        self.aligner = aligner
        #call super
        super(DecisionTreeNode, self).__init__()
        
    def is_end_node(self):
        #TODO: tokens recalculated 
        # better to do this with iteration
        # This can then be written without if statements
        len_wit_a = len(self.aligner.witness_a.tokens())
        if self.pointer_a == len_wit_a:
            return True
        len_wit_b = len(self.aligner.witness_b.tokens())
        if self.pointer_b == len_wit_b:
            return True
        return False
        
class DecisionTree(AStar):
    def __init__(self, aligner):
        self.aligner = aligner
        
    def create_childnodes(self):
#         # check whether a token is a match
#         token_a = tokens_a[self.state.pointer_a]
#         token_b = tokens_b[self.state.pointer_b]
#         match = token_a.eql(token_b)
# 
#         #move this to above!
#         if not match:
#             # we try to handle mismatch as an omission
#             self.states
#             
        pass

    

class Aligner(object):
    '''
    Decision Tree based aligner
    This is a prototype: 
    There are limitations
    At first it only works with just two witnesses
    It works with two pointers.
    The pointers start at the beginning (first token) of each of the witnesses.
    If this prototype works, one of the witnesses has to be replaced by the graph
    Then if that works, the enhanced suffix array based optimalization has to be integrated.
    Also we start by tracking one decision state.. we should keep track of more state
    to reach the right conclusion. The A* algorithm works very well for that.
    '''


    def __init__(self, witness_a, witness_b):
        '''
        Constructor
        '''
        self.witness_a = witness_a
        self.witness_b = witness_b
        self.tree = DecisionTree(self)
        
    def align(self):
        '''
        Every step we have 3 choices:
        1) Move pointer witness a --> omission
        2) Move pointer witness b --> addition
        3) Move pointer of both witness a/b  --> match
        Note: a replacement is omission followed by an addition or the other way around
        
        Choice 1 and 2 are only possible if token a and b are not a match OR when tokens are repeated.
        For now I ignore token repetition..
        '''
        # extract tokens from witness (note that this can be done in a streaming manner if desired)
        tokens_a = self.witness_a.tokens()
        tokens_b = self.witness_b.tokens()
        # create virtual decision tree (nodes are created on demand)
        # see above
        # create start node
        start = DecisionTreeNode(self)
        
        # search the decision tree
        result = self.tree.search(start)
        print(result)
        
        
        pass
    


'''
Created on Nov 20, 2014

@author: Tara L Andrews
'''

import unittest
from collatex import Collation
from collatex.core_functions import collate_pretokenized_json


class Test(unittest.TestCase):
    def testPlainWitness(self):
        plain_witness = {'id': 'A', 'content': 'The quick brown fox jumped over the lazy dogs.'}
        c = Collation()
        c.add_witness(plain_witness)
        self.assertEqual(len(c.witnesses[0].tokens()), 10)

    def testPretokenizedWitnessAdd(self):
        pt_witness = {
                    "id": "A",
                    "tokens": [
                        {"t": "A", "ref": 123},
                        {"t": "black and blue", "adj": True},
                        {"t": "cat", "id": "xyz"},
                        {"t": "bird", "id": "abc"}
                    ]
                }
        c = Collation()
        c.add_witness(pt_witness)
        self.assertEqual(len(c.witnesses[0].tokens()), 4)

    def testPretokenizedWitness(self):
        pretokenized_witness = {
            "witnesses": [
                {
                    "id": "A",
                    "tokens": [
                        {"t": "A", "ref": 123},
                        {"t": "black", "adj": True},
                        {"t": "cat", "id": "xyz"},
                        {"t": "bird", "id": "abc"}
                    ]
                },
                {
                    "id": "B",
                    "tokens": [
                        {"t": "A"},
                        {"t": "white", "adj": True},
                        {"t": "mousedog bird", "adj": False}
                    ]
                }
            ]
        }
        result = collate_pretokenized_json(pretokenized_witness)
        self.assertEqual(len(result.rows[0].to_list()), 4)
        self.assertEqual(len(result.rows[1].to_list()), 4)
        # The second witness should have a token that reads 'mousedog bird'.
        self.assertIn("mousedog bird", result.rows[1].to_list())


if __name__ == '__main__':
    unittest.main()



'''
Created on Jun 8, 2014

@author: Ronald Haentjens Dekker
'''
import unittest
from tests import unit_disabled
from collatex import Collation
from collatex import collate


class Test(unittest.TestCase):

    @unit_disabled
    # TODO This aligns 'the cat' instead of 'is'!
    def testDoubleTransposition1(self):
        collation = Collation()
        collation.add_plain_witness("A", "the cat is black")
        collation.add_plain_witness("B", "black is the cat")
        alignment_table = collate(collation)
        self.assertEquals(["the cat", "is", "black"], alignment_table.rows[0].to_list())
        self.assertEquals(["black", "is", "the cat"], alignment_table.rows[1].to_list())

    def testDoubleTransposition2(self):
        # Either the 'a' can align or the 'b' can. See also #3 below.
        collation = Collation()
        collation.add_plain_witness("A", "a b")
        collation.add_plain_witness("B", "b a")
        alignment_table = collate(collation)
        witness_a_list = alignment_table.rows[0].to_list()
        self.assertEquals(len(witness_a_list), 3)
        witness_b_list = alignment_table.rows[1].to_list()
        self.assertEquals(len(witness_b_list), 3)
        matching_tokens = []
        for idx in range(3):
            if witness_a_list[idx] == witness_b_list[idx]:
                matching_tokens.append(witness_a_list[idx])
        self.assertEquals(len(matching_tokens), 1)

    def testDoubleTransposition3(self):
        # Tricky. Aligning a and c can work; so can aligning b and c. Both
        # are equally valid, and both can crop up.
        # Let's test that each row has four values, and that two of the
        # columns have identical values, and that c is one of those columns.
        collation = Collation()
        collation.add_plain_witness("A", "a b c")
        collation.add_plain_witness("B", "b a c")
        alignment_table = collate(collation)
        witness_a_list = alignment_table.rows[0].to_list()
        self.assertEquals(len(witness_a_list), 4)
        witness_b_list = alignment_table.rows[1].to_list()
        self.assertEquals(len(witness_b_list), 4)
        matching_tokens = []
        for idx in range(4):
            if witness_a_list[idx] == witness_b_list[idx]:
                matching_tokens.append(witness_a_list[idx])
        self.assertEquals(len(matching_tokens), 2)
        self.assertIn("c", matching_tokens)

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
Created on Apr 27, 2014

@author: Ronald Haentjens Dekker
'''
import unittest
from tests import unit_disabled
from ClusterShell.RangeSet import RangeSet
from array import array
from collatex import Collation
from collatex.collatex_suffix import ExtendedSuffixArray, Block
from collatex.suffix_based_scorer import Scorer




class Test(unittest.TestCase):
    def assertIntervalIn(self, start, length, nr_of_occurrences, intervals):
        found = False
        for lcp_interval in intervals:
            if lcp_interval.token_start_position == start and lcp_interval.minimum_block_length == length and lcp_interval.number_of_occurrences == nr_of_occurrences:
                found = True
                break 
        if not found:
            self.fail("Interval with "+str(start)+" and "+str(length)+" and "+str(nr_of_occurrences)+" not found in "+str(intervals))
    
    def test_combined_string_hermans_case(self):
        collation = Collation()
        collation.add_plain_witness("W1", "a b c d F g h i ! K ! q r s t")
        collation.add_plain_witness("W2", "a b c d F g h i ! q r s t")
        # $ is meant to separate witnesses here
        self.assertEquals("a b c d F g h i ! K ! q r s t $1 a b c d F g h i ! q r s t", collation.get_combined_string())
    
    # test whether the witness->range mapping works
    @unit_disabled
    def test_witness_ranges_hermans_case(self):
        collation = Collation()
        collation.add_plain_witness("W1", "a b c d F g h i ! K ! q r s t")
        collation.add_plain_witness("W2", "a b c d F g h i ! q r s t")
        self.assertEquals(RangeSet("0-14"), collation.get_range_for_witness("W1"))
        self.assertEquals(RangeSet("16-28"), collation.get_range_for_witness("W2"))

# TODO: re-enable test!    
    # Note: LCP intervals can overlap
    @unit_disabled
    def test_lcp_intervals_failing_use_case_old_algorithm(self):
        collation = Collation()
        collation.add_plain_witness("W1", "the cat and the dog")
        collation.add_plain_witness("W2", "the dog and the cat")
        parent_lcp_intervals, child_lcp_intervals = collation.get_lcp_intervals()
        self.assertIn((1,2), parent_lcp_intervals)
        self.assertIn((3,4), parent_lcp_intervals)
        self.assertIn((5,6), parent_lcp_intervals)
        self.assertIn((7, 10), parent_lcp_intervals)
        self.assertIn((7,8), child_lcp_intervals[7])
        self.assertIn((9,10), child_lcp_intervals[7])
    
# TODO: re-enable test!
    @unit_disabled
    def test_lcp_child_intervals_darwin(self):
        lcp_array = array('i', [0, 0, 0, 96, 96, 0, 151, 9, 1, 105, 105, 0, 83, 83, 0, 95, 95, 0, 39, 39, 0, 24, 24, 0, 108, 1, 0, 232, 32, 0, 181, 39, 0, 185, 43, 1, 33, 33, 0, 159, 17, 0, 160, 18, 0, 106, 106, 0, 60, 60, 0, 171, 29, 1, 215, 15, 1, 122, 15, 1, 57, 57, 1, 153, 11, 1, 165, 23, 0, 9, 9, 1, 170, 28, 0, 214, 14, 0, 62, 62, 1, 191, 49, 0, 61, 61, 0, 148, 6, 1, 8, 8, 0, 19, 19, 0, 123, 16, 0, 75, 75, 1, 90, 90, 1, 28, 28, 0, 167, 25, 1, 112, 5, 1, 132, 25, 0, 50, 50, 0, 31, 31, 0, 77, 77, 0, 97, 97, 0, 6, 6, 0, 38, 38, 0, 63, 63, 0, 30, 30, 0, 80, 80, 0, 154, 12, 0, 145, 3, 0, 129, 22, 3, 67, 67, 0, 88, 88, 0, 45, 45, 0, 217, 17, 1, 22, 22, 0, 166, 24, 0, 25, 25, 0, 201, 1, 0, 155, 13, 1, 120, 13, 0, 175, 33, 0, 195, 53, 0, 135, 28, 0, 10, 10, 0, 147, 5, 0, 138, 31, 0, 161, 19, 1, 73, 73, 0, 198, 1, 0, 86, 86, 0, 74, 74, 1, 111, 4, 0, 210, 10, 0, 84, 84, 0, 42, 42, 0, 199, 2, 0, 119, 12, 0, 46, 46, 1, 202, 2, 0, 71, 71, 0, 40, 40, 0, 142, 0, 0, 52, 52, 0, 168, 26, 2, 113, 6, 1, 163, 21, 0, 133, 26, 0, 3, 3, 0, 186, 44, 1, 101, 101, 0, 193, 51, 2, 227, 27, 0, 107, 1, 37, 37, 1, 140, 33, 0, 1, 205, 5, 0, 47, 47, 0, 127, 20, 1, 65, 65, 0, 230, 30, 0, 41, 41, 0, 91, 91, 0, 1, 1, 0, 200, 3, 0, 156, 14, 0, 0, 76, 76, 0, 109, 2, 1, 182, 40, 0, 68, 68, 1, 14, 14, 0, 126, 19, 0, 34, 34, 1, 192, 50, 1, 85, 85, 1, 128, 21, 2, 66, 66, 1, 183, 41, 1, 220, 20, 1, 5, 5, 1, 212, 12, 2, 174, 32, 2, 226, 26, 1, 59, 59, 0, 16, 16, 0, 218, 18, 0, 23, 23, 1, 11, 11, 0, 36, 36, 1, 178, 36, 0, 51, 51, 0, 213, 13, 1, 190, 48, 0, 2, 2, 1, 222, 22, 1, 188, 46, 0, 78, 78, 0, 53, 53, 0, 197, 0, 0, 136, 29, 1, 219, 19, 1, 12, 12, 0, 114, 7, 0, 89, 89, 0, 172, 30, 2, 216, 16, 0, 21, 21, 0, 209, 9, 0, 81, 81, 0, 102, 102, 0, 134, 27, 0, 98, 98, 0, 131, 24, 0, 4, 4, 0, 35, 35, 0, 179, 37, 0, 224, 24, 0, 82, 82, 0, 72, 72, 0, 139, 32, 0, 125, 18, 0, 103, 103, 0, 121, 14, 0, 189, 47, 0, 184, 42, 0, 7, 7, 1, 17, 17, 0, 207, 7, 0, 221, 21, 0, 20, 20, 0, 196, 54, 0, 79, 79, 1, 204, 4, 1, 144, 2, 2, 94, 94, 1, 56, 56, 0, 211, 11, 1, 194, 52, 3, 228, 28, 1, 157, 15, 1, 69, 69, 1, 54, 54, 1, 115, 8, 1, 173, 31, 1, 225, 25, 1, 177, 35, 1, 100, 100, 0, 203, 3, 0, 150, 8, 0, 104, 104, 0, 143, 1, 1, 93, 93, 0, 118, 11, 0, 29, 29, 1, 64, 64, 1, 146, 4, 1, 137, 30, 1, 229, 29, 2, 70, 70, 1, 44, 44, 1, 49, 49, 1, 117, 10, 0, 152, 10, 0, 130, 23, 1, 26, 26, 1, 110, 3, 1, 158, 16, 0, 124, 17, 0, 206, 6, 0, 141, 34, 1, 92, 92, 0, 32, 32, 1, 27, 27, 0, 58, 58, 0, 162, 20, 0, 13, 13, 0, 187, 45, 1, 223, 23, 0, 43, 43, 0, 48, 48, 0, 176, 34, 0, 99, 99, 0, 149, 7, 1, 231, 31, 1, 180, 38, 0, 18, 18, 0, 55, 55, 0, 169, 27, 2, 164, 22, 1, 208, 8, 1, 116, 9, 0, 87, 87, 0, 15, 15])
        collation = Collation()
        _, child_lcp_intervals = collation.get_lcp_intervals(lcp=lcp_array)
        self.assertEqual([(5, 7), (8, 10)], child_lcp_intervals[5])
        self.assertEqual([(513, 515),(516, 518),(519, 521),(522, 524),(525,527)], child_lcp_intervals[513])

# TODO: re-enable test!
    @unit_disabled
    def test_lcp_child_intervals_hermans_case(self):
        collation = Collation()
        collation.add_plain_witness("W1", "a b c d F g h i ! K ! q r s t")
        collation.add_plain_witness("W2", "a b c d F g h i ! q r s t")
        collation.add_plain_witness("W3", "a b c d E g h i ! q r s t")
        _, child_lcp_intervals = collation.get_lcp_intervals()
        self.assertFalse(child_lcp_intervals)

    @unit_disabled
    def test_non_overlapping_blocks_black_cat(self):
        collation = Collation()
        collation.add_plain_witness("W1", "the black cat")
        collation.add_plain_witness("W2", "the black cat")
        algorithm = Scorer(collation)
        blocks = algorithm._get_non_overlapping_repeating_blocks()
        block1 = Block(RangeSet("0-2, 4-6"))
        self.assertEqual([block1], blocks)

    #TODO: Fix number of siblings!
    @unit_disabled
    def test_blocks_failing_transposition_use_case_old_algorithm(self):
        collation = Collation()
        collation.add_plain_witness("W1", "the cat and the dog")
        collation.add_plain_witness("W2", "the dog and the cat")
        algorithm = Scorer(collation)
        blocks = algorithm._get_non_overlapping_repeating_blocks()
        block1 = Block(RangeSet("0-1, 9-10"))
        block2 = Block(RangeSet("3-4, 6-7"))
        block3 = Block(RangeSet("2, 8"))
        self.assertEqual([block1, block2, block3], blocks)

    @unit_disabled
    def test_non_overlapping_blocks_Hermans(self):
        collation = Collation()
        collation.add_plain_witness("W1", "a b c d F g h i ! K ! q r s t")
        collation.add_plain_witness("W2", "a b c d F g h i ! q r s t")
        algorithm = Scorer(collation)
        blocks = algorithm._get_non_overlapping_repeating_blocks()
        self.assertIn(Block(RangeSet("0-8, 16-24")), blocks) # a b c d F g h i !
        self.assertIn(Block(RangeSet("11-14, 25-28")), blocks) # q r s t

    @unit_disabled
    def test_blocks_Hermans_case_three_witnesses(self):
        collation = Collation()
        collation.add_plain_witness("W1", "a b c d F g h i ! K ! q r s t")
        collation.add_plain_witness("W2", "a b c d F g h i ! q r s t")
        collation.add_plain_witness("W3", "a b c d E g h i ! q r s t")
        algorithm = Scorer(collation)
        blocks = algorithm._get_non_overlapping_repeating_blocks()
        self.assertIn(Block(RangeSet("0-3, 16-19, 30-33")), blocks) # a b c d
        self.assertIn(Block(RangeSet("5-7, 21-23, 35-37")), blocks) # g h i
        self.assertIn(Block(RangeSet("10-14, 24-28, 38-42")), blocks) # ! q r s t
        self.assertIn(Block(RangeSet("4, 20")), blocks) # F
        

    # In the new approach nothing should be split
    @unit_disabled
    def test_blocks_splitting_token_case(self):
        collation = Collation()
        collation.add_plain_witness("W1", "a c b c")
        collation.add_plain_witness("W2", "a c b")
        algorithm = Scorer(collation)
        blocks = algorithm._get_non_overlapping_repeating_blocks()
        block1 = Block(RangeSet("0-2, 5-7")) # a c b
        self.assertIn(block1, blocks)

    @unit_disabled
    def test_block_witnesses_Hermans_case_two_witnesses(self):
        collation = Collation()
        collation.add_plain_witness("W1", "a b c d F g h i ! K ! q r s t")
        collation.add_plain_witness("W2", "a b c d F g h i ! q r s t")
        algorithm = Scorer(collation)
        block_witness = algorithm._get_block_witness(collation.witnesses[0])
        self.assertEquals(["a b c d F g h i !", "q r s t"], block_witness.debug())
        block_witness = algorithm._get_block_witness(collation.witnesses[1])
        self.assertEquals(["a b c d F g h i !", "q r s t"], block_witness.debug())

    @unit_disabled
    def test_block_witnesses_Hermans_case(self):
        collation = Collation()
        collation.add_plain_witness("W1", "a b c d F g h i ! K ! q r s t")
        collation.add_plain_witness("W2", "a b c d F g h i ! q r s t")
        collation.add_plain_witness("W3", "a b c d E g h i ! q r s t")
        algorithm = Scorer(collation)
        block_witness1 = algorithm._get_block_witness(collation.witnesses[0])
        self.assertEquals(["a b c d", "F", "g h i", "! q r s t"], block_witness1.debug())
        block_witness2 = algorithm._get_block_witness(collation.witnesses[1])
        self.assertEquals(["a b c d", "F", "g h i", "! q r s t"], block_witness2.debug())
        block_witness3 = algorithm._get_block_witness(collation.witnesses[2])
        self.assertEquals(["a b c d", "g h i", "! q r s t"], block_witness3.debug())
        
    def test_filter_potential_blocks(self):
        collation = Collation()
        collation.add_plain_witness("W1", "a a")
        collation.add_plain_witness("w2", "a")
        extsufarr = collation.to_extended_suffix_array()
        potential_blocks = extsufarr.split_lcp_array_into_intervals()
        algorithm = Scorer(collation)
        algorithm.filter_potential_blocks(potential_blocks)
        self.assertFalse(potential_blocks)
    
    # LCP interval is not ascending nor descending
    @unit_disabled
    def test_split_lcp_intervals_into_smaller_intervals(self):
        collation = Collation()
        collation.add_plain_witness("W1", "the cat")
        collation.add_plain_witness("W2", "the cat")
        collation.add_plain_witness("W3", "the cat")
        extsufarr = collation.to_extended_suffix_array()
        split_intervals = extsufarr.split_lcp_array_into_intervals()
        self.assertIntervalIn(0, 2, 3, split_intervals) # the cat
        self.assertIntervalIn(1, 1, 3, split_intervals) # cat
        self.assertEqual(2, len(split_intervals), "More items: "+str(split_intervals))
        
    # LCP interval is ascending
    @unit_disabled
    def test_split_lcp_intervals_into_smaller_intervals_2(self):
        collation = Collation()
        collation.add_plain_witness("W1", "the")
        collation.add_plain_witness("W2", "the cat")
        collation.add_plain_witness("W3", "the cat sits")
        extsufarr = collation.to_extended_suffix_array()
        split_intervals = extsufarr.split_lcp_array_into_intervals()
        self.assertIntervalIn(0, 1, 3, split_intervals) # the
        self.assertIntervalIn(2, 2, 2, split_intervals) # the cat
        self.assertIntervalIn(3, 1, 2, split_intervals) # cat
        self.assertEqual(3, len(split_intervals), "More items: "+str(split_intervals))

    # LCP interval is descending
    def test_split_lcp_intervals_descending_LCP(self):
        lcp_array = array('i', [0, 20, 20, 20, 4])
        sa_array = array('i', [0, 1, 2, 3, 4]) # FAKED!
        extsuffarr = ExtendedSuffixArray(None, sa_array, lcp_array)
        split_intervals = extsuffarr.split_lcp_array_into_intervals()
        self.assertIntervalIn(0, 20, 4, split_intervals)
        self.assertIntervalIn(0, 4, 5, split_intervals)
        self.assertEqual(2, len(split_intervals), "More items: "+str(split_intervals))
        
    # LCP interval is first ascending, then descending
    def test_split_lcp_intervals_ascending_then_descending_LCP(self):
        lcp_array = array('i', [0, 10, 149, 93, 7, 1])
        sa_array = array('i', [0, 1, 2, 3, 4, 5]) # FAKED!
        extsuffarr = ExtendedSuffixArray(None, sa_array, lcp_array)
        split_intervals = extsuffarr.split_lcp_array_into_intervals()
        self.assertIntervalIn(0, 10, 4, split_intervals)
        self.assertIntervalIn(1, 149, 2, split_intervals)
        self.assertIntervalIn(1, 93, 3, split_intervals)
        self.assertIntervalIn(0, 7, 5, split_intervals)
        self.assertIntervalIn(0, 1, 6, split_intervals)
        self.assertEqual(5, len(split_intervals), "More items: "+str(split_intervals))
        
    def test_split_lcp_intervals_ascending_descending_ascending(self):
        lcp_array =  array('i', [0, 4, 143, 87, 1, 1, 12, 93, 93, 37])
        sa_array = array('i', [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]) # FAKED!
        extsuffarr = ExtendedSuffixArray(None, sa_array, lcp_array)
        split_intervals = extsuffarr.split_lcp_array_into_intervals()
        self.assertIntervalIn(1, 143, 2, split_intervals)
        self.assertIntervalIn(1, 87, 3, split_intervals)
        self.assertIntervalIn(0, 4, 4, split_intervals)
        self.assertIntervalIn(6, 93, 3, split_intervals)
        self.assertIntervalIn(0, 1, 10, split_intervals)
        self.assertIntervalIn(5, 12, 5, split_intervals)
        self.assertIntervalIn(6, 37, 4, split_intervals)

    @unit_disabled
    def test_filter_potential_blocks(self):
        collation = Collation()
        collation.add_plain_witness("W1", "the fox jumps over the fox")
        collation.add_plain_witness("w2", "the fox jumps over the dog")
        potential_blocks = collation.calculate_potential_blocks()
        collation.filter_potential_blocks(potential_blocks)
        self.fail("TESTING!")

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
Created on Nov 13, 2014

@author: ronald
'''
from collatex.core_classes import create_table_visualization

# optionally load the IPython dependencies
try:
    from IPython.display import HTML
    from IPython.display import SVG
    from IPython.core.display import display
    # import graphviz python library
    from graphviz import Digraph
except:
    pass


# create visualization of alignment table
def display_alignment_table_as_HTML(at):
    prettytable = create_table_visualization(at)
    html = prettytable.get_html_string(formatting=True)
    return display(HTML(html))

# visualize the variant graph into SVG format
# from networkx.drawing.nx_agraph import to_agraph
def display_variant_graph_as_SVG(graph):
        # create new Digraph
        dot = Digraph(format="svg", graph_attr={'rankdir': 'LR'})
        # add nodes
        for n,nodedata in graph.graph.nodes(data=True):
            dot.node(str(n), nodedata["label"])
        # add edges
        for u,v,edgedata in graph.graph.edges_iter(data=True):
            dot.edge(str(u), str(v), edgedata["label"])
        # render the dot graph to SVG
        # Note: this creates a file
        svg = dot.render()
        # display using the IPython SVG module
        return display(SVG(svg))


# def in_ipython():
#     try:
#         get_ipython().config  # @UndefinedVariable
# #         print('Called by IPython.')
#         return True
#     except:
#         return False





# display alignment table 
#     if in_ipython():
#         return display(JSON(json))
#     if in_ipython():
#         print(json)
#         return


# DISPLAY PART OF THE VARIANT_GRAPH IN SVG!
#     and in_ipython:
#         # visualize the variant graph into SVG format
#         from networkx.drawing.nx_agraph import to_agraph
#         agraph = to_agraph(graph.graph)
#         svg = agraph.draw(format="svg", prog="dot", args="-Grankdir=LR -Gid=VariantGraph")
#         return display(SVG(svg)) 



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Ronald Haentjens Dekker'
__email__ = 'ronald.dekker@huygens.knaw.nl'
__version__ = '2.0.0pre10'

from collatex.core_functions import Collation
from collatex.core_functions import collate
from collatex.core_functions import collate_pretokenized_json

__all__ = ["Collation", "collate", "collate_pretokenized_json"]





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

setup(
    name='collatex',
    version='2.0.0pre12',
    description='CollateX is a collation tool.',
    long_description=readme + '\n\n' + history,
    author='Ronald Haentjens Dekker',
    author_email='ronald.dekker@huygens.knaw.nl',
    url='https://github.com/rhdekker/collatex',
    packages=[
        'collatex','ClusterShell'
    ],
    package_dir={'collatex':
                 'collatex'},
    include_package_data=True,
    install_requires=['networkx','prettytable'
    ],
    license="GPLv3",
    zip_safe=False,
    keywords='CollateX',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
    ],
    test_suite='tests',
)



'''
Created on Apr 19, 2014

@author: Ronald Haentjens Dekker

This module defines the core collation concepts of CollateX

Tokenizer, Witness, VariantGraph, CollationAlgorithm
'''
import networkx as nx
from _collections import deque
from networkx.algorithms.dag import topological_sort
import re
from prettytable import PrettyTable
from textwrap import fill
from collatex.exceptions import TokenError

class Row(object):
    
    def __init__(self, header):
        self.cells = []
        self.header = header
    
    def append(self, cell):
        self.cells.append(cell)
        
    def to_list(self):
        return self.cells
    

class Column(object):
    
    def __init__(self):
        self.tokens_per_witness = {}
        self.variant = False  

    def put(self, sigil, token):
        self.tokens_per_witness[sigil]=token
                
    pass


class AlignmentTable(object):
    
    def __init__(self, collation, graph=None, layout="horizontal"):
        self.collation = collation
        self.graph = graph
        self.layout = layout
        self.columns = []
        self.rows = []
        if graph:
            self._construct_table()

    def _construct_table(self):
        ranking = VariantGraphRanking.of(self.graph)
        vertices_per_rank = ranking.byRank
        # construct columns        
        for rank in vertices_per_rank:
            column = None
            vertices = vertices_per_rank[rank]
            for vertex in vertices:
                if vertex == self.graph.start or vertex == self.graph.end:
                    continue
                if column == None:
                    column = Column()
                    self.columns.append(column)
                #find incoming edges for this vertex and check their labels
                edges = self.graph.in_edges(vertex, data=True)
                for (_, _, attrs) in edges:
                    sigli = attrs["label"]
                    for sigil in sigli.split(", "):
                        vertex_attrs = self.graph.graph.node[vertex]
                        token = vertex_attrs["label"]
                        column.put(sigil, token)
                # set status: is a column variant or invariant
                column.variant = len(vertices) > 1 or len(column.tokens_per_witness) != len(self.collation.witnesses)

        # construct rows
        for witness in self.collation.witnesses:
            sigil = witness.sigil
            # we create a new Row for every witness
            row = Row(sigil)
            self.rows.append(row)
            for column in self.columns:
                if sigil in column.tokens_per_witness:
                    row.append(column.tokens_per_witness[sigil])
                else:
                    #TODO: Nil would be nicer here
                    # since the dash is part of the visualization
                    row.append("-")
    
    def __str__(self, *args, **kwargs):
        return str(create_table_visualization(self))
                    
# DISPLAY PART OF THE VARIANT GRAPH IN PLAIN/HTML AND VERTICAL OR HORIZONTAL!
def create_table_visualization(table):
    # create visualization of alignment table
    if table.layout == "vertical":    
        prettytable = visualizeTableVertically(table)
    elif table.layout == "horizontal":
        prettytable = visualizeTableHorizontal(table)
    else:
        raise Exception("Unknown table layout: "+table.layout)
    return prettytable

def visualizeTableHorizontal(table):
    # print the table horizontal
    x = PrettyTable()
    x.header=False
    for row in table.rows:
        cells = [row.header]
        cells.extend(row.cells)
        x.add_row(cells)
    # alignment can only be set after the field names are known.
    # since add_row sets the field names, it has to be set after x.add_row(cells)
    x.align="l"
    return x

def visualizeTableVertically(table):
    # print the table vertically
    x = PrettyTable()
    x.hrules = 1
    for row in table.rows:
        x.add_column(row.header, [fill(cell, 20) for cell in row.cells])
    return x

   
# not used in the suffix implementation
# Tokenizer inside suffix array library is used
class Tokenizer(object):
    
    #by default the tokenizer splits on space characters    
    def tokenize(self, contents):
        return contents.split()

class WordPunctuationTokenizer(object):
    #tokenizer splits on punctuation or whitespace
    def tokenize(self, contents):
#       the remarked regular expression keeps the whitespace 
#       return re.findall("[.?!,;:]+[\\s]*|[^.?!,;:\\s]+[\\s]*", contents)
        return re.findall(r'\w+|[^\w\s]+', contents)

class Token(object):
    # tokendata comes in the dictionary that we use for JSON input.
    def __init__(self, tokendata=None):
        if tokendata is None:
            # We can have empty tokens.
            self.token_string = ""
            self.token_data = {}
        elif 'n' in tokendata:
            self.token_string = tokendata['n']
        elif 't' in tokendata:
            self.token_string = tokendata['t']
        else:
            raise TokenError('No defined token string in tokendata')
        self.token_data = tokendata

    def __repr__(self):
        return self.token_string

class Witness(object):
    
    def __init__(self, witnessdata):
        self.sigil = witnessdata['id']
        self._tokens = []
        if 'content' in witnessdata:
            self.content = witnessdata['content']
            # print("Witness "+sigil+" TOKENIZER IS CALLED!")
            tokenizer = WordPunctuationTokenizer()
            tokens_as_strings = tokenizer.tokenize(self.content)
            for token_string in tokens_as_strings:
                self._tokens.append(Token({'t':token_string}))
        elif 'tokens' in witnessdata:
            for tk in witnessdata['tokens']:
                self._tokens.append(Token(tk))
            # TODO no idea what this content string is needed for.
            self.content = ' '.join([x.token_string for x in self._tokens])
            
    def tokens(self):
        return self._tokens

class VariantGraph(object):
    
    def __init__(self):
        self.graph = nx.DiGraph()
        self.start = self.add_vertex(Token())
        self.end = self.add_vertex(Token())
    
#     def is_directed(self):
#         return self.graph.is_directed()
#     
    # vertex creation uses a unique ID, since the token_content does not have to be unique   
    # we store the token content in the label 
    def add_vertex(self, token):
        '''
        :type token: Token
        '''
        node_id = self.graph.number_of_nodes()
        #print("Adding node: "+node_id+":"+token_content)
        self.graph.add_node(node_id, label= token.token_string)
        return node_id
    
    def connect(self, source, target, witnesses):
        """
        :type source: integer
        :type target: integer
        """
        #print("Adding Edge: "+source+":"+target)
        if self.graph.has_edge(source, target):
            self.graph[source][target]["label"]+=", "+str(witnesses)
        else:    
            self.graph.add_edge(source, target, label=witnesses)
        
    def remove_edge(self, source, target):
        self.graph.remove_edge(source, target)
        
    def remove_node(self, node):
        self.graph.remove_node(node)      
        
    def vertices(self):
        return self.graph.nodes()
    
    def edges(self):
        return self.graph.edges()
    
    def edge_between(self, node, node2):
        #return self.graph.get_edge_data(node, node2)
        return self.graph.has_edge(node, node2)
    
    def in_edges(self, node, data=False):
        return self.graph.in_edges(nbunch=node, data=data)
    
    def out_edges(self, node, data=False):
        return self.graph.out_edges(nbunch=node, data=data)
    
    def vertex_attributes(self, node):
        return self.graph.node[node]
  
    # Note: generator implementation
    def vertexWith(self, content):
        try:
            vertex_to_find = (n for n in self.graph if self.graph.node[n]['label']==content).next()
            return vertex_to_find    
        except StopIteration:
            raise Exception("Vertex with "+content+" not found!")    
  
class CollationAlgorithm(object):
    def merge(self, graph, witness_sigil, witness_tokens, alignments = {}):  
        """
        :type graph: VariantGraph
        """
        #NOTE: token_to_vertex only contains newly generated vertices
        token_to_vertex = {}
        last = graph.start
        for token in witness_tokens:
            vertex = alignments.get(token, None)
            if vertex == None:
                vertex = graph.add_vertex(token)
                token_to_vertex[token] = vertex
            #TODO: add token to vertex, for example: vertex.add(token)
            graph.connect(last, vertex, witness_sigil)
            last = vertex
        graph.connect(last, graph.end, witness_sigil)
        return token_to_vertex

#TODO: define abstract collation class


'''
 This function joins the variant graph in place.
 This function is a straight port of the Java version of CollateX.
    :type graph: VariantGraph
 TODO: add transposition support!   
'''
def join(graph):
    processed = set()
    end = graph.end
    queue = deque()
    for (_, neighbor) in graph.out_edges(graph.start):
        queue.appendleft(neighbor)
    while queue:
        vertex = queue.popleft()
        out_edges = graph.out_edges(vertex)
        if len(out_edges) is 1:
            (_, join_candidate) = out_edges[0]
            can_join = join_candidate != end and len(graph.in_edges(join_candidate))==1
            if can_join:
                # Note: since there is no normalized/non normalized content in the graph
                # a space character is added here for non punctuation tokens
                label = graph.vertex_attributes(join_candidate)["label"]
                if re.match(r'^\W', label):
                    graph.vertex_attributes(vertex)["label"]+=label
                else:
                    graph.vertex_attributes(vertex)["label"]+=" "+label
                for (_, neighbor, data) in graph.out_edges(join_candidate, data=True):
                    graph.remove_edge(join_candidate, neighbor)
                    graph.connect(vertex, neighbor, data["label"])
                graph.remove_edge(vertex, join_candidate)
                graph.remove_node(join_candidate) 
                queue.appendleft(vertex);
                continue;
        processed.add(vertex)
        for (_, neighbor) in out_edges:
            # FIXME: Why do we run out of memory in some cases here, if this is not checked?
            if not neighbor in processed:
                queue.appendleft(neighbor)
                
# Port of VariantGraphRanking class from Java
# This is a minimal port; only bare bones
class VariantGraphRanking(object):
    # Do not class the constructor, use the of class method instead!
    def __init__(self):
        #Note: A vertex can have only one rank
        #however, a rank can be assigned to multiple vertices
        self.byVertex = {}
        self.byRank = {}
        
    def apply(self, vertex):
        return self.byVertex[vertex]
    
    @classmethod
    def of(cls, graph):
        variant_graph_ranking = VariantGraphRanking()                
        topological_sorted_vertices = topological_sort(graph.graph)
        for v in topological_sorted_vertices:
            rank = -1
            for (source, _) in graph.in_edges(v):
                rank = max(rank, variant_graph_ranking.byVertex[source])
            rank += 1
            variant_graph_ranking.byVertex[v]=rank
            variant_graph_ranking.byRank.setdefault(rank, []).append(v)
        return variant_graph_ranking

    
    




